@@ -34,6 +34,9 @@
 #pragma GCC poison PETSC_PETSC_HAVE_MEMMOVE
 #pragma GCC poison PETSC_PETSC_HAVE_MKSTEMP
 #pragma GCC poison PETSC_PETSC_HAVE_MPIUNI
+#pragma GCC poison PETSC_PETSC_HAVE_MPI_COUNT
+#pragma GCC poison PETSC_PETSC_HAVE_MPI_LARGE_COUNT
+#pragma GCC poison PETSC_PETSC_HAVE_MPI_REDUCE_LOCAL
 #pragma GCC poison PETSC_PETSC_HAVE_OPENBLAS
 #pragma GCC poison PETSC_PETSC_HAVE_OPENBLAS_SET_NUM_THREADS
 #pragma GCC poison PETSC_PETSC_HAVE_OPENMP
@@ -42,7 +45,6 @@
 #pragma GCC poison PETSC_PETSC_HAVE_POPEN
 #pragma GCC poison PETSC_PETSC_HAVE_RAND
 #pragma GCC poison PETSC_PETSC_HAVE_REAL___FP16
-#pragma GCC poison PETSC_PETSC_HAVE_REGEX
 #pragma GCC poison PETSC_PETSC_HAVE_SETJMP_H
 #pragma GCC poison PETSC_PETSC_HAVE_SETLASTERROR
 #pragma GCC poison PETSC_PETSC_HAVE_SLEEP
