@@ -11,7 +11,7 @@
 */
 
 #include <petsc/private/pcasmimpl.h> /*I "petscpc.h" I*/
-#include "petsc/private/matimpl.h"
+#include <petsc/private/matimpl.h>
 
 static PetscErrorCode PCView_ASM(PC pc, PetscViewer viewer)
 {
@@ -43,15 +43,15 @@ static PetscErrorCode PCView_ASM(PC pc, PetscViewer viewer)
         PetscCall(PetscViewerASCIIPrintf(viewer, "  Use -%sksp_view ::ascii_info_detail to display information for all blocks\n", prefix ? prefix : ""));
         PetscCall(PetscViewerGetSubViewer(viewer, PETSC_COMM_SELF, &sviewer));
         if (rank == 0) {
-          PetscCall(PetscViewerASCIIPushTab(viewer));
+          PetscCall(PetscViewerASCIIPushTab(sviewer));
           PetscCall(KSPView(osm->ksp[0], sviewer));
-          PetscCall(PetscViewerASCIIPopTab(viewer));
+          PetscCall(PetscViewerASCIIPopTab(sviewer));
         }
         PetscCall(PetscViewerRestoreSubViewer(viewer, PETSC_COMM_SELF, &sviewer));
       }
     } else {
       PetscCall(PetscViewerASCIIPushSynchronized(viewer));
-      PetscCall(PetscViewerASCIISynchronizedPrintf(viewer, "  [%d] number of local blocks = %" PetscInt_FMT "\n", (int)rank, osm->n_local_true));
+      PetscCall(PetscViewerASCIISynchronizedPrintf(viewer, "  [%d] number of local blocks = %" PetscInt_FMT "\n", rank, osm->n_local_true));
       PetscCall(PetscViewerFlush(viewer));
       PetscCall(PetscViewerASCIIPrintf(viewer, "  Local solver information for each block is in the following KSP and PC objects:\n"));
       PetscCall(PetscViewerASCIIPushTab(viewer));
@@ -59,13 +59,12 @@ static PetscErrorCode PCView_ASM(PC pc, PetscViewer viewer)
       PetscCall(PetscViewerGetSubViewer(viewer, PETSC_COMM_SELF, &sviewer));
       for (i = 0; i < osm->n_local_true; i++) {
         PetscCall(ISGetLocalSize(osm->is[i], &bsz));
-        PetscCall(PetscViewerASCIISynchronizedPrintf(sviewer, "[%d] local block number %" PetscInt_FMT ", size = %" PetscInt_FMT "\n", (int)rank, i, bsz));
+        PetscCall(PetscViewerASCIIPrintf(sviewer, "[%d] local block number %" PetscInt_FMT ", size = %" PetscInt_FMT "\n", rank, i, bsz));
         PetscCall(KSPView(osm->ksp[i], sviewer));
-        PetscCall(PetscViewerASCIISynchronizedPrintf(sviewer, "- - - - - - - - - - - - - - - - - -\n"));
+        PetscCall(PetscViewerASCIIPrintf(sviewer, "- - - - - - - - - - - - - - - - - -\n"));
       }
       PetscCall(PetscViewerRestoreSubViewer(viewer, PETSC_COMM_SELF, &sviewer));
       PetscCall(PetscViewerASCIIPopTab(viewer));
-      PetscCall(PetscViewerFlush(viewer));
       PetscCall(PetscViewerASCIIPopSynchronized(viewer));
     }
   } else if (isstring) {
@@ -153,16 +152,17 @@ static PetscErrorCode PCASMPrintSubdomains(PC pc)
 
 static PetscErrorCode PCSetUp_ASM(PC pc)
 {
-  PC_ASM     *osm = (PC_ASM *)pc->data;
-  PetscBool   flg;
-  PetscInt    i, m, m_local;
-  MatReuse    scall = MAT_REUSE_MATRIX;
-  IS          isl;
-  KSP         ksp;
-  PC          subpc;
-  const char *prefix, *pprefix;
-  Vec         vec;
-  DM         *domain_dm = NULL;
+  PC_ASM       *osm = (PC_ASM *)pc->data;
+  PetscBool     flg;
+  PetscInt      i, m, m_local;
+  MatReuse      scall = MAT_REUSE_MATRIX;
+  IS            isl;
+  KSP           ksp;
+  PC            subpc;
+  const char   *prefix, *pprefix;
+  Vec           vec;
+  DM           *domain_dm = NULL;
+  MatNullSpace *nullsp    = NULL;
 
   PetscFunctionBegin;
   if (!pc->setupcalled) {
@@ -204,7 +204,7 @@ static PetscErrorCode PCSetUp_ASM(PC pc)
 
       inwork.max = osm->n_local_true;
       inwork.sum = osm->n_local_true;
-      PetscCall(MPIU_Allreduce(&inwork, &outwork, 1, MPIU_2INT, MPIU_MAXSUM_OP, PetscObjectComm((PetscObject)pc)));
+      PetscCallMPI(MPIU_Allreduce(&inwork, &outwork, 1, MPIU_2INT, MPIU_MAXSUM_OP, PetscObjectComm((PetscObject)pc)));
       osm->n_local = outwork.max;
       osm->n       = outwork.sum;
 
@@ -277,6 +277,7 @@ static PetscErrorCode PCSetUp_ASM(PC pc)
        Destroy the blocks from the previous iteration
     */
     if (pc->flag == DIFFERENT_NONZERO_PATTERN) {
+      PetscCall(MatGetNullSpaces(osm->n_local_true, osm->pmat, &nullsp));
       PetscCall(MatDestroyMatrices(osm->n_local_true, &osm->pmat));
       scall = MAT_INITIAL_MATRIX;
     }
@@ -284,6 +285,7 @@ static PetscErrorCode PCSetUp_ASM(PC pc)
 
   /* Destroy previous submatrices of a different type than pc->pmat since MAT_REUSE_MATRIX won't work in that case */
   if (scall == MAT_REUSE_MATRIX && osm->sub_mat_type) {
+    PetscCall(MatGetNullSpaces(osm->n_local_true, osm->pmat, &nullsp));
     if (osm->n_local_true > 0) PetscCall(MatDestroySubMatrices(osm->n_local_true, &osm->pmat));
     scall = MAT_INITIAL_MATRIX;
   }
@@ -295,11 +297,12 @@ static PetscErrorCode PCSetUp_ASM(PC pc)
   if (scall == MAT_INITIAL_MATRIX) {
     PetscCall(PetscObjectGetOptionsPrefix((PetscObject)pc->pmat, &pprefix));
     for (i = 0; i < osm->n_local_true; i++) PetscCall(PetscObjectSetOptionsPrefix((PetscObject)osm->pmat[i], pprefix));
+    if (nullsp) PetscCall(MatRestoreNullSpaces(osm->n_local_true, osm->pmat, &nullsp));
   }
 
   /* Convert the types of the submatrices (if needbe) */
   if (osm->sub_mat_type) {
-    for (i = 0; i < osm->n_local_true; i++) PetscCall(MatConvert(osm->pmat[i], osm->sub_mat_type, MAT_INPLACE_MATRIX, &(osm->pmat[i])));
+    for (i = 0; i < osm->n_local_true; i++) PetscCall(MatConvert(osm->pmat[i], osm->sub_mat_type, MAT_INPLACE_MATRIX, &osm->pmat[i]));
   }
 
   if (!pc->setupcalled) {
@@ -348,7 +351,7 @@ static PetscErrorCode PCSetUp_ASM(PC pc)
       PetscCall(VecScatterCreate(osm->ly, isll, osm->y[i], isl, &osm->lrestriction[i]));
       PetscCall(ISDestroy(&isll));
       PetscCall(ISDestroy(&isl));
-      if (osm->lprolongation) { /* generate a scatter from y[i] to ly picking only the the non-overlapping is_local[i] entries */
+      if (osm->lprolongation) { /* generate a scatter from y[i] to ly picking only the non-overlapping is_local[i] entries */
         ISLocalToGlobalMapping ltog;
         IS                     isll, isll_local;
         const PetscInt        *idx_local;
@@ -566,6 +569,7 @@ static PetscErrorCode PCApplyTranspose_ASM(PC pc, Vec x, Vec y)
   ScatterMode forward = SCATTER_FORWARD, reverse = SCATTER_REVERSE;
 
   PetscFunctionBegin;
+  PetscCheck(osm->n_local_true <= 1 || osm->loctype == PC_COMPOSITE_ADDITIVE, PetscObjectComm((PetscObject)pc), PETSC_ERR_SUP, "Not yet implemented");
   /*
      Support for limiting the restriction or interpolation to only local
      subdomain values (leaving the other values 0).
@@ -596,10 +600,10 @@ static PetscErrorCode PCApplyTranspose_ASM(PC pc, Vec x, Vec y)
   /* do the local solves */
   for (i = 0; i < n_local_true; ++i) {
     /* solve the overlapping i-block */
-    PetscCall(PetscLogEventBegin(PC_ApplyOnBlocks, osm->ksp[i], osm->x[i], osm->y[i], 0));
+    PetscCall(PetscLogEventBegin(PC_ApplyTransposeOnBlocks, osm->ksp[i], osm->x[i], osm->y[i], 0));
     PetscCall(KSPSolveTranspose(osm->ksp[i], osm->x[i], osm->y[i]));
     PetscCall(KSPCheckSolve(osm->ksp[i], pc, osm->y[i]));
-    PetscCall(PetscLogEventEnd(PC_ApplyOnBlocks, osm->ksp[i], osm->x[i], osm->y[i], 0));
+    PetscCall(PetscLogEventEnd(PC_ApplyTransposeOnBlocks, osm->ksp[i], osm->x[i], osm->y[i], 0));
 
     if (osm->lprolongation && osm->type != PC_ASM_RESTRICT) { /* interpolate the non-overlapping i-block solution to the local solution */
       PetscCall(VecScatterBegin(osm->lprolongation[i], osm->y[i], osm->ly, ADD_VALUES, forward));
@@ -646,7 +650,7 @@ static PetscErrorCode PCReset_ASM(PC pc)
     PetscCall(PetscFree(osm->x));
     PetscCall(PetscFree(osm->y));
   }
-  PetscCall(PCASMDestroySubdomains(osm->n_local_true, osm->is, osm->is_local));
+  PetscCall(PCASMDestroySubdomains(osm->n_local_true, &osm->is, &osm->is_local));
   PetscCall(ISDestroy(&osm->lis));
   PetscCall(VecDestroy(&osm->lx));
   PetscCall(VecDestroy(&osm->ly));
@@ -686,7 +690,7 @@ static PetscErrorCode PCDestroy_ASM(PC pc)
   PetscFunctionReturn(PETSC_SUCCESS);
 }
 
-static PetscErrorCode PCSetFromOptions_ASM(PC pc, PetscOptionItems *PetscOptionsObject)
+static PetscErrorCode PCSetFromOptions_ASM(PC pc, PetscOptionItems PetscOptionsObject)
 {
   PC_ASM         *osm = (PC_ASM *)pc->data;
   PetscInt        blocks, ovl;
@@ -741,7 +745,7 @@ static PetscErrorCode PCASMSetLocalSubdomains_ASM(PC pc, PetscInt n, IS is[], IS
     if (is_local) {
       for (i = 0; i < n; i++) PetscCall(PetscObjectReference((PetscObject)is_local[i]));
     }
-    PetscCall(PCASMDestroySubdomains(osm->n_local_true, osm->is, osm->is_local));
+    PetscCall(PCASMDestroySubdomains(osm->n_local_true, &osm->is, &osm->is_local));
 
     if (osm->ksp && osm->n_local_true != n) {
       for (i = 0; i < osm->n_local_true; i++) PetscCall(KSPDestroy(&osm->ksp[i]));
@@ -785,7 +789,7 @@ static PetscErrorCode PCASMSetTotalSubdomains_ASM(PC pc, PetscInt N, IS *is, IS
 
   PetscFunctionBegin;
   PetscCheck(N >= 1, PetscObjectComm((PetscObject)pc), PETSC_ERR_ARG_OUTOFRANGE, "Number of total blocks must be > 0, N = %" PetscInt_FMT, N);
-  PetscCheck(!is && !is_local, PetscObjectComm((PetscObject)pc), PETSC_ERR_SUP, "Use PCASMSetLocalSubdomains() to set specific index sets\n\they cannot be set globally yet.");
+  PetscCheck(!is && !is_local, PetscObjectComm((PetscObject)pc), PETSC_ERR_SUP, "Use PCASMSetLocalSubdomains() to set specific index sets, they cannot be set globally yet.");
 
   /*
      Split the subdomains equally among all processors
@@ -793,10 +797,10 @@ static PetscErrorCode PCASMSetTotalSubdomains_ASM(PC pc, PetscInt N, IS *is, IS
   PetscCallMPI(MPI_Comm_rank(PetscObjectComm((PetscObject)pc), &rank));
   PetscCallMPI(MPI_Comm_size(PetscObjectComm((PetscObject)pc), &size));
   n = N / size + ((N % size) > rank);
-  PetscCheck(n, PETSC_COMM_SELF, PETSC_ERR_ARG_OUTOFRANGE, "Process %d must have at least one block: total processors %d total blocks %" PetscInt_FMT, (int)rank, (int)size, N);
+  PetscCheck(n, PETSC_COMM_SELF, PETSC_ERR_ARG_OUTOFRANGE, "Process %d must have at least one block: total processors %d total blocks %" PetscInt_FMT, rank, size, N);
   PetscCheck(!pc->setupcalled || n == osm->n_local_true, PETSC_COMM_SELF, PETSC_ERR_ARG_WRONGSTATE, "PCASMSetTotalSubdomains() should be called before PCSetUp().");
   if (!pc->setupcalled) {
-    PetscCall(PCASMDestroySubdomains(osm->n_local_true, osm->is, osm->is_local));
+    PetscCall(PCASMDestroySubdomains(osm->n_local_true, &osm->is, &osm->is_local));
 
     osm->n_local_true = n;
     osm->is           = NULL;
@@ -901,7 +905,7 @@ static PetscErrorCode PCASMSetSubMatType_ASM(PC pc, MatType sub_mat_type)
   PetscFunctionReturn(PETSC_SUCCESS);
 }
 
-/*@C
+/*@
   PCASMSetLocalSubdomains - Sets the local subdomains (for this processor only) for the additive Schwarz preconditioner `PCASM`.
 
   Collective
@@ -909,10 +913,11 @@ static PetscErrorCode PCASMSetSubMatType_ASM(PC pc, MatType sub_mat_type)
   Input Parameters:
 + pc       - the preconditioner context
 . n        - the number of subdomains for this processor (default value = 1)
-. is       - the index set that defines the subdomains for this processor
-         (or `NULL` for PETSc to determine subdomains)
-- is_local - the index sets that define the local part of the subdomains for this processor, not used unless PCASMType is PC_ASM_RESTRICT
-         (or `NULL` to not provide these)
+. is       - the index set that defines the subdomains for this processor (or `NULL` for PETSc to determine subdomains)
+             the values of the `is` array are copied so you can free the array (not the `IS` in the array) after this call
+- is_local - the index sets that define the local part of the subdomains for this processor, not used unless `PCASMType` is `PC_ASM_RESTRICT`
+             (or `NULL` to not provide these). The values of the `is_local` array are copied so you can free the array
+             (not the `IS` in the array) after this call
 
   Options Database Key:
 . -pc_asm_local_blocks <blks> - Sets number of local blocks
@@ -920,19 +925,19 @@ static PetscErrorCode PCASMSetSubMatType_ASM(PC pc, MatType sub_mat_type)
   Level: advanced
 
   Notes:
-  The `IS` numbering is in the parallel, global numbering of the vector for both is and is_local
+  The `IS` numbering is in the parallel, global numbering of the vector for both `is` and `is_local`
 
   By default the `PCASM` preconditioner uses 1 block per processor.
 
   Use `PCASMSetTotalSubdomains()` to set the subdomains for all processors.
 
-  If is_local is provided and `PCASMType` is `PC_ASM_RESTRICT` then the solution only over the is_local region is interpolated
-  back to form the global solution (this is the standard restricted additive Schwarz method)
+  If `is_local` is provided and `PCASMType` is `PC_ASM_RESTRICT` then the solution only over the `is_local` region is interpolated
+  back to form the global solution (this is the standard restricted additive Schwarz method, RASM)
 
-  If the is_local is provided and `PCASMType` is `PC_ASM_INTERPOLATE` or `PC_ASM_NONE` then an error is generated since there is
+  If `is_local` is provided and `PCASMType` is `PC_ASM_INTERPOLATE` or `PC_ASM_NONE` then an error is generated since there is
   no code to handle that case.
 
-.seealso: `PCASM`, `PCASMSetTotalSubdomains()`, `PCASMSetOverlap()`, `PCASMGetSubKSP()`,
+.seealso: [](ch_ksp), `PCASM`, `PCASMSetTotalSubdomains()`, `PCASMSetOverlap()`, `PCASMGetSubKSP()`,
           `PCASMCreateSubdomains2D()`, `PCASMGetLocalSubdomains()`, `PCASMType`, `PCASMSetType()`, `PCGASM`
 @*/
 PetscErrorCode PCASMSetLocalSubdomains(PC pc, PetscInt n, IS is[], IS is_local[])
@@ -943,7 +948,7 @@ PetscErrorCode PCASMSetLocalSubdomains(PC pc, PetscInt n, IS is[], IS is_local[]
   PetscFunctionReturn(PETSC_SUCCESS);
 }
 
-/*@C
+/*@
   PCASMSetTotalSubdomains - Sets the subdomains for all processors for the
   additive Schwarz preconditioner, `PCASM`.
 
@@ -952,10 +957,10 @@ PetscErrorCode PCASMSetLocalSubdomains(PC pc, PetscInt n, IS is[], IS is_local[]
   Input Parameters:
 + pc       - the preconditioner context
 . N        - the number of subdomains for all processors
-. is       - the index sets that define the subdomains for all processors
-         (or `NULL` to ask PETSc to determine the subdomains)
-- is_local - the index sets that define the local part of the subdomains for this processor
-         (or `NULL` to not provide this information)
+. is       - the index sets that define the subdomains for all processors (or `NULL` to ask PETSc to determine the subdomains)
+             the values of the `is` array are copied so you can free the array (not the `IS` in the array) after this call
+- is_local - the index sets that define the local part of the subdomains for this processor (or `NULL` to not provide this information)
+             The values of the `is_local` array are copied so you can free the array (not the `IS` in the array) after this call
 
   Options Database Key:
 . -pc_asm_blocks <blks> - Sets total blocks
@@ -963,7 +968,7 @@ PetscErrorCode PCASMSetLocalSubdomains(PC pc, PetscInt n, IS is[], IS is_local[]
   Level: advanced
 
   Notes:
-  Currently you cannot use this to set the actual subdomains with the argument is or is_local.
+  Currently you cannot use this to set the actual subdomains with the argument `is` or `is_local`.
 
   By default the `PCASM` preconditioner uses 1 block per processor.
 
@@ -974,7 +979,7 @@ PetscErrorCode PCASMSetLocalSubdomains(PC pc, PetscInt n, IS is[], IS is_local[]
 
   The `IS` numbering is in the parallel, global numbering of the vector for both is and is_local
 
-.seealso: `PCASM`, `PCASMSetLocalSubdomains()`, `PCASMSetOverlap()`, `PCASMGetSubKSP()`,
+.seealso: [](ch_ksp), `PCASM`, `PCASMSetLocalSubdomains()`, `PCASMSetOverlap()`, `PCASMGetSubKSP()`,
           `PCASMCreateSubdomains2D()`, `PCGASM`
 @*/
 PetscErrorCode PCASMSetTotalSubdomains(PC pc, PetscInt N, IS is[], IS is_local[])
@@ -1021,7 +1026,7 @@ PetscErrorCode PCASMSetTotalSubdomains(PC pc, PetscInt N, IS is[], IS is_local[]
   `PCASMSetOverlap()` merely allows PETSc to extend that overlap further
   if desired.
 
-.seealso: `PCASM`, `PCASMSetTotalSubdomains()`, `PCASMSetLocalSubdomains()`, `PCASMGetSubKSP()`,
+.seealso: [](ch_ksp), `PCASM`, `PCASMSetTotalSubdomains()`, `PCASMSetLocalSubdomains()`, `PCASMGetSubKSP()`,
           `PCASMCreateSubdomains2D()`, `PCASMGetLocalSubdomains()`, `MatIncreaseOverlap()`, `PCGASM`
 @*/
 PetscErrorCode PCASMSetOverlap(PC pc, PetscInt ovl)
@@ -1058,7 +1063,7 @@ PetscErrorCode PCASMSetOverlap(PC pc, PetscInt ovl)
   if the is_local arguments are passed to `PCASMSetLocalSubdomains()` then they are used when `PC_ASM_RESTRICT` has been selected
   to limit the local processor interpolation
 
-.seealso: `PCASM`, `PCASMSetTotalSubdomains()`, `PCASMGetSubKSP()`,
+.seealso: [](ch_ksp), `PCASM`, `PCASMSetTotalSubdomains()`, `PCASMGetSubKSP()`,
           `PCASMCreateSubdomains2D()`, `PCASMType`, `PCASMSetLocalType()`, `PCASMGetLocalType()`, `PCGASM`
 @*/
 PetscErrorCode PCASMSetType(PC pc, PCASMType type)
@@ -1093,7 +1098,7 @@ PetscErrorCode PCASMSetType(PC pc, PCASMType type)
 
   Level: intermediate
 
-.seealso: `PCASM`, `PCASMSetTotalSubdomains()`, `PCASMGetSubKSP()`, `PCGASM`,
+.seealso: [](ch_ksp), `PCASM`, `PCASMSetTotalSubdomains()`, `PCASMGetSubKSP()`, `PCGASM`,
           `PCASMCreateSubdomains2D()`, `PCASMType`, `PCASMSetType()`, `PCASMSetLocalType()`, `PCASMGetLocalType()`
 @*/
 PetscErrorCode PCASMGetType(PC pc, PCASMType *type)
@@ -1122,7 +1127,7 @@ PetscErrorCode PCASMGetType(PC pc, PCASMType *type)
 
   Level: intermediate
 
-.seealso: `PCASM`, `PCASMSetType()`, `PCASMGetType()`, `PCASMGetLocalType()`, `PCASMType`, `PCCompositeType`
+.seealso: [](ch_ksp), `PCASM`, `PCASMSetType()`, `PCASMGetType()`, `PCASMGetLocalType()`, `PCASMType`, `PCCompositeType`
 @*/
 PetscErrorCode PCASMSetLocalType(PC pc, PCCompositeType type)
 {
@@ -1153,7 +1158,7 @@ PetscErrorCode PCASMSetLocalType(PC pc, PCCompositeType type)
 
   Level: intermediate
 
-.seealso: `PCASM`, `PCASMSetType()`, `PCASMGetType()`, `PCASMSetLocalType()`, `PCASMCreate()`, `PCASMType`, `PCCompositeType`
+.seealso: [](ch_ksp), `PCASM`, `PCASMSetType()`, `PCASMGetType()`, `PCASMSetLocalType()`, `PCASMType`, `PCCompositeType`
 @*/
 PetscErrorCode PCASMGetLocalType(PC pc, PCCompositeType *type)
 {
@@ -1175,7 +1180,7 @@ PetscErrorCode PCASMGetLocalType(PC pc, PCCompositeType *type)
 
   Level: intermediate
 
-.seealso: `PCASM`, `PCASMSetLocalSubdomains()`, `PCASMSetTotalSubdomains()`, `PCASMGetSubKSP()`,
+.seealso: [](ch_ksp), `PCASM`, `PCASMSetLocalSubdomains()`, `PCASMSetTotalSubdomains()`, `PCASMGetSubKSP()`,
           `PCASMCreateSubdomains2D()`
 @*/
 PetscErrorCode PCASMSetSortIndices(PC pc, PetscBool doSort)
@@ -1197,9 +1202,8 @@ PetscErrorCode PCASMSetSortIndices(PC pc, PetscBool doSort)
 . pc - the preconditioner context
 
   Output Parameters:
-+ n_local     - the number of blocks on this processor or NULL
-. first_local - the global number of the first block on this processor or NULL,
-                 all processors must request or all must pass NULL
++ n_local     - the number of blocks on this processor or `NULL`
+. first_local - the global number of the first block on this processor or `NULL`, all processors must request or all must pass `NULL`
 - ksp         - the array of `KSP` contexts
 
   Level: advanced
@@ -1209,10 +1213,7 @@ PetscErrorCode PCASMSetSortIndices(PC pc, PetscBool doSort)
 
   You must call `KSPSetUp()` before calling `PCASMGetSubKSP()`.
 
-  Fortran Notes:
-  The output argument 'ksp' must be an array of sufficient length or `PETSC_NULL_KSP`. The latter can be used to learn the necessary length.
-
-.seealso: `PCASM`, `PCASMSetTotalSubdomains()`, `PCASMSetOverlap()`,
+.seealso: [](ch_ksp), `PCASM`, `PCASMSetTotalSubdomains()`, `PCASMSetOverlap()`,
           `PCASMCreateSubdomains2D()`,
 @*/
 PetscErrorCode PCASMGetSubKSP(PC pc, PetscInt *n_local, PetscInt *first_local, KSP *ksp[])
@@ -1225,37 +1226,34 @@ PetscErrorCode PCASMGetSubKSP(PC pc, PetscInt *n_local, PetscInt *first_local, K
 
 /*MC
    PCASM - Use the (restricted) additive Schwarz method, each block is (approximately) solved with
-           its own `KSP` object.
+           its own `KSP` object, {cite}`dryja1987additive` and {cite}`1sbg`
 
    Options Database Keys:
-+  -pc_asm_blocks <blks> - Sets total blocks. Defaults to one block per MPI rank.
-.  -pc_asm_overlap <ovl> - Sets overlap
++  -pc_asm_blocks <blks>                          - Sets total blocks. Defaults to one block per MPI process.
+.  -pc_asm_overlap <ovl>                          - Sets overlap
 .  -pc_asm_type [basic,restrict,interpolate,none] - Sets `PCASMType`, default is restrict. See `PCASMSetType()`
--  -pc_asm_local_type [additive, multiplicative] - Sets `PCCompositeType`, default is additive. See `PCASMSetLocalType()`
+.  -pc_asm_dm_subdomains <bool>                   - use subdomains defined by the `DM` with `DMCreateDomainDecomposition()`
+-  -pc_asm_local_type [additive, multiplicative]  - Sets `PCCompositeType`, default is additive. See `PCASMSetLocalType()`
 
    Level: beginner
 
    Notes:
    If you run with, for example, 3 blocks on 1 processor or 3 blocks on 3 processors you
-   will get a different convergence rate due to the default option of -pc_asm_type restrict. Use
-    -pc_asm_type basic to get the same convergence behavior
+   will get a different convergence rate due to the default option of `-pc_asm_type restrict`. Use
+   `-pc_asm_type basic` to get the same convergence behavior
 
    Each processor can have one or more blocks, but a block cannot be shared by more
    than one processor. Use `PCGASM` for subdomains shared by multiple processes.
 
-   To set options on the solvers for each block append -sub_ to all the `KSP`, and `PC`
-   options database keys. For example, -sub_pc_type ilu -sub_pc_factor_levels 1 -sub_ksp_type preonly
+   To set options on the solvers for each block append `-sub_` to all the `KSP`, and `PC`
+   options database keys. For example, `-sub_pc_type ilu -sub_pc_factor_levels 1 -sub_ksp_type preonly`
 
    To set the options on the solvers separate for each block call `PCASMGetSubKSP()`
    and set the options directly on the resulting `KSP` object (you can access its `PC` with `KSPGetPC()`)
 
-    References:
-+   * - M Dryja, OB Widlund, An additive variant of the Schwarz alternating method for the case of many subregions
-     Courant Institute, New York University Technical report
--   * - Barry Smith, Petter Bjorstad, and William Gropp, Domain Decompositions: Parallel Multilevel Methods for Elliptic Partial Differential Equations,
-    Cambridge University Press.
+   If the `PC` has an associated `DM`, then, by default, `DMCreateDomainDecomposition()` is used to create the subdomains
 
-.seealso: `PCCreate()`, `PCSetType()`, `PCType`, `PC`, `PCASMType`, `PCCompositeType`,
+.seealso: [](ch_ksp), `PCCreate()`, `PCSetType()`, `PCType`, `PC`, `PCASMType`, `PCCompositeType`,
           `PCBJACOBI`, `PCASMGetSubKSP()`, `PCASMSetLocalSubdomains()`, `PCASMType`, `PCASMGetType()`, `PCASMSetLocalType()`, `PCASMGetLocalType()`
           `PCASMSetTotalSubdomains()`, `PCSetModifySubMatrices()`, `PCASMSetOverlap()`, `PCASMSetType()`, `PCCompositeType`
 M*/
@@ -1332,10 +1330,7 @@ PETSC_EXTERN PetscErrorCode PCCreate_ASM(PC pc)
   This generates nonoverlapping subdomains; the `PCASM` will generate the overlap
   from these if you use `PCASMSetLocalSubdomains()`
 
-  Fortran Notes:
-  You must provide the array outis[] already allocated of length n.
-
-.seealso: `PCASM`, `PCASMSetLocalSubdomains()`, `PCASMDestroySubdomains()`
+.seealso: [](ch_ksp), `PCASM`, `PCASMSetLocalSubdomains()`, `PCASMDestroySubdomains()`
 @*/
 PetscErrorCode PCASMCreateSubdomains(Mat A, PetscInt n, IS *outis[])
 {
@@ -1497,23 +1492,26 @@ PetscErrorCode PCASMCreateSubdomains(Mat A, PetscInt n, IS *outis[])
 
   Level: advanced
 
-.seealso: `PCASM`, `PCASMCreateSubdomains()`, `PCASMSetLocalSubdomains()`
+  Developer Note:
+  The `IS` arguments should be a *[]
+
+.seealso: [](ch_ksp), `PCASM`, `PCASMCreateSubdomains()`, `PCASMSetLocalSubdomains()`
 @*/
-PetscErrorCode PCASMDestroySubdomains(PetscInt n, IS is[], IS is_local[])
+PetscErrorCode PCASMDestroySubdomains(PetscInt n, IS *is[], IS *is_local[])
 {
   PetscInt i;
 
   PetscFunctionBegin;
   if (n <= 0) PetscFunctionReturn(PETSC_SUCCESS);
-  if (is) {
-    PetscAssertPointer(is, 2);
-    for (i = 0; i < n; i++) PetscCall(ISDestroy(&is[i]));
-    PetscCall(PetscFree(is));
+  if (*is) {
+    PetscAssertPointer(*is, 2);
+    for (i = 0; i < n; i++) PetscCall(ISDestroy(&(*is)[i]));
+    PetscCall(PetscFree(*is));
   }
-  if (is_local) {
-    PetscAssertPointer(is_local, 3);
-    for (i = 0; i < n; i++) PetscCall(ISDestroy(&is_local[i]));
-    PetscCall(PetscFree(is_local));
+  if (is_local && *is_local) {
+    PetscAssertPointer(*is_local, 3);
+    for (i = 0; i < n; i++) PetscCall(ISDestroy(&(*is_local)[i]));
+    PetscCall(PetscFree(*is_local));
   }
   PetscFunctionReturn(PETSC_SUCCESS);
 }
@@ -1544,13 +1542,10 @@ PetscErrorCode PCASMDestroySubdomains(PetscInt n, IS is[], IS is_local[])
   preconditioners.  More general related routines are
   `PCASMSetTotalSubdomains()` and `PCASMSetLocalSubdomains()`.
 
-  Fortran Notes:
-  The `IS` must be declared as an array of length long enough to hold `Nsub` entries
-
-.seealso: `PCASM`, `PCASMSetTotalSubdomains()`, `PCASMSetLocalSubdomains()`, `PCASMGetSubKSP()`,
+.seealso: [](ch_ksp), `PCASM`, `PCASMSetTotalSubdomains()`, `PCASMSetLocalSubdomains()`, `PCASMGetSubKSP()`,
           `PCASMSetOverlap()`
 @*/
-PetscErrorCode PCASMCreateSubdomains2D(PetscInt m, PetscInt n, PetscInt M, PetscInt N, PetscInt dof, PetscInt overlap, PetscInt *Nsub, IS **is, IS **is_local)
+PetscErrorCode PCASMCreateSubdomains2D(PetscInt m, PetscInt n, PetscInt M, PetscInt N, PetscInt dof, PetscInt overlap, PetscInt *Nsub, IS *is[], IS *is_local[])
 {
   PetscInt i, j, height, width, ystart, xstart, yleft, yright, xleft, xright, loc_outer;
   PetscInt nidx, *idx, loc, ii, jj, count;
@@ -1625,7 +1620,7 @@ PetscErrorCode PCASMCreateSubdomains2D(PetscInt m, PetscInt n, PetscInt M, Petsc
   Note:
   The `IS` numbering is in the parallel, global numbering of the vector.
 
-.seealso: `PCASM`, `PCASMSetTotalSubdomains()`, `PCASMSetOverlap()`, `PCASMGetSubKSP()`,
+.seealso: [](ch_ksp), `PCASM`, `PCASMSetTotalSubdomains()`, `PCASMSetOverlap()`, `PCASMGetSubKSP()`,
           `PCASMCreateSubdomains2D()`, `PCASMSetLocalSubdomains()`, `PCASMGetLocalSubmatrices()`
 @*/
 PetscErrorCode PCASMGetLocalSubdomains(PC pc, PetscInt *n, IS *is[], IS *is_local[])
@@ -1666,7 +1661,7 @@ PetscErrorCode PCASMGetLocalSubdomains(PC pc, PetscInt *n, IS *is[], IS *is_loca
 
   Usually one would use `PCSetModifySubMatrices()` to change the submatrices in building the preconditioner.
 
-.seealso: `PCASM`, `PCASMSetTotalSubdomains()`, `PCASMSetOverlap()`, `PCASMGetSubKSP()`,
+.seealso: [](ch_ksp), `PCASM`, `PCASMSetTotalSubdomains()`, `PCASMSetOverlap()`, `PCASMGetSubKSP()`,
           `PCASMCreateSubdomains2D()`, `PCASMSetLocalSubdomains()`, `PCASMGetLocalSubdomains()`, `PCSetModifySubMatrices()`
 @*/
 PetscErrorCode PCASMGetLocalSubmatrices(PC pc, PetscInt *n, Mat *mat[])
@@ -1701,7 +1696,7 @@ PetscErrorCode PCASMGetLocalSubmatrices(PC pc, PetscInt *n, Mat *mat[])
 - flg - boolean indicating whether to use subdomains defined by the `DM`
 
   Options Database Key:
-. -pc_asm_dm_subdomains <bool> - use subdomains defined by the `DM`
+. -pc_asm_dm_subdomains <bool> - use subdomains defined by the `DM` with `DMCreateDomainDecomposition()`
 
   Level: intermediate
 
@@ -1709,7 +1704,10 @@ PetscErrorCode PCASMGetLocalSubmatrices(PC pc, PetscInt *n, Mat *mat[])
   `PCASMSetTotalSubdomains()` and `PCASMSetOverlap()` take precedence over `PCASMSetDMSubdomains()`,
   so setting either of the first two effectively turns the latter off.
 
-.seealso: `PCASM`, `PCASMGetDMSubdomains()`, `PCASMSetTotalSubdomains()`, `PCASMSetOverlap()`
+  Developer Note:
+  This should be `PCASMSetUseDMSubdomains()`, similarly for the options database key
+
+.seealso: [](ch_ksp), `PCASM`, `PCASMGetDMSubdomains()`, `PCASMSetTotalSubdomains()`, `PCASMSetOverlap()`
           `PCASMCreateSubdomains2D()`, `PCASMSetLocalSubdomains()`, `PCASMGetLocalSubdomains()`
 @*/
 PetscErrorCode PCASMSetDMSubdomains(PC pc, PetscBool flg)
@@ -1739,7 +1737,10 @@ PetscErrorCode PCASMSetDMSubdomains(PC pc, PetscBool flg)
 
   Level: intermediate
 
-.seealso: `PCASM`, `PCASMSetDMSubdomains()`, `PCASMSetTotalSubdomains()`, `PCASMSetOverlap()`
+  Developer Note:
+  This should be `PCASMSetUseDMSubdomains()`
+
+.seealso: [](ch_ksp), `PCASM`, `PCASMSetDMSubdomains()`, `PCASMSetTotalSubdomains()`, `PCASMSetOverlap()`
           `PCASMCreateSubdomains2D()`, `PCASMSetLocalSubdomains()`, `PCASMGetLocalSubdomains()`
 @*/
 PetscErrorCode PCASMGetDMSubdomains(PC pc, PetscBool *flg)
@@ -1756,7 +1757,7 @@ PetscErrorCode PCASMGetDMSubdomains(PC pc, PetscBool *flg)
   PetscFunctionReturn(PETSC_SUCCESS);
 }
 
-/*@C
+/*@
   PCASMGetSubMatType - Gets the matrix type used for `PCASM` subsolves, as a string.
 
   Not Collective
@@ -1769,7 +1770,7 @@ PetscErrorCode PCASMGetDMSubdomains(PC pc, PetscBool *flg)
 
   Level: advanced
 
-.seealso: `PCASM`, `PCASMSetSubMatType()`, `PCSetType()`, `VecSetType()`, `MatType`, `Mat`
+.seealso: [](ch_ksp), `PCASM`, `PCASMSetSubMatType()`, `PCSetType()`, `VecSetType()`, `MatType`, `Mat`
 @*/
 PetscErrorCode PCASMGetSubMatType(PC pc, MatType *sub_mat_type)
 {
@@ -1779,7 +1780,7 @@ PetscErrorCode PCASMGetSubMatType(PC pc, MatType *sub_mat_type)
   PetscFunctionReturn(PETSC_SUCCESS);
 }
 
-/*@C
+/*@
   PCASMSetSubMatType - Set the type of matrix used for `PCASM` subsolves
 
   Collective
@@ -1797,7 +1798,7 @@ PetscErrorCode PCASMGetSubMatType(PC pc, MatType *sub_mat_type)
 
   Level: advanced
 
-.seealso: `PCASM`, `PCASMGetSubMatType()`, `PCSetType()`, `VecSetType()`, `MatType`, `Mat`
+.seealso: [](ch_ksp), `PCASM`, `PCASMGetSubMatType()`, `PCSetType()`, `VecSetType()`, `MatType`, `Mat`
 @*/
 PetscErrorCode PCASMSetSubMatType(PC pc, MatType sub_mat_type)
 {
