@@ -4,62 +4,7 @@
   much of anything.
 */
 
-#include <petsc/private/matimpl.h> /*I "petscmat.h" I*/
-#include <petsc/private/vecimpl.h>
-
-struct _MatShellOps {
-  /*  3 */ PetscErrorCode (*mult)(Mat, Vec, Vec);
-  /*  5 */ PetscErrorCode (*multtranspose)(Mat, Vec, Vec);
-  /* 17 */ PetscErrorCode (*getdiagonal)(Mat, Vec);
-  /* 43 */ PetscErrorCode (*copy)(Mat, Mat, MatStructure);
-  /* 60 */ PetscErrorCode (*destroy)(Mat);
-};
-
-struct _n_MatShellMatFunctionList {
-  PetscErrorCode (*symbolic)(Mat, Mat, Mat, void **);
-  PetscErrorCode (*numeric)(Mat, Mat, Mat, void *);
-  PetscErrorCode (*destroy)(void *);
-  MatProductType ptype;
-  char          *composedname; /* string to identify routine with double dispatch */
-  char          *resultname;   /* result matrix type */
-
-  struct _n_MatShellMatFunctionList *next;
-};
-typedef struct _n_MatShellMatFunctionList *MatShellMatFunctionList;
-
-typedef struct {
-  struct _MatShellOps ops[1];
-
-  /* The user will manage the scaling and shifts for the MATSHELL, not the default */
-  PetscBool managescalingshifts;
-
-  /* support for MatScale, MatShift and MatMultAdd */
-  PetscScalar vscale, vshift;
-  Vec         dshift;
-  Vec         left, right;
-  Vec         left_work, right_work;
-  Vec         left_add_work, right_add_work;
-
-  /* support for MatAXPY */
-  Mat              axpy;
-  PetscScalar      axpy_vscale;
-  Vec              axpy_left, axpy_right;
-  PetscObjectState axpy_state;
-
-  /* support for ZeroRows/Columns operations */
-  IS         zrows;
-  IS         zcols;
-  Vec        zvals;
-  Vec        zvals_w;
-  VecScatter zvals_sct_r;
-  VecScatter zvals_sct_c;
-
-  /* MatMat operations */
-  MatShellMatFunctionList matmat;
-
-  /* user defined context */
-  PetscContainer ctxcontainer;
-} Mat_Shell;
+#include <../src/mat/impls/shell/shell.h> /*I "petscmat.h" I*/
 
 /*
      Store and scale values on zeroed rows
@@ -142,7 +87,7 @@ static PetscErrorCode MatShellPostZeroRight(Mat A, Vec x)
 /*
       xx = diag(left)*x
 */
-static PetscErrorCode MatShellPreScaleLeft(Mat A, Vec x, Vec *xx)
+static PetscErrorCode MatShellPreScaleLeft(Mat A, Vec x, Vec *xx, PetscBool conjugate)
 {
   Mat_Shell *shell = (Mat_Shell *)A->data;
 
@@ -152,7 +97,19 @@ static PetscErrorCode MatShellPreScaleLeft(Mat A, Vec x, Vec *xx)
     *xx = x;
   } else {
     if (!shell->left_work) PetscCall(VecDuplicate(shell->left, &shell->left_work));
-    PetscCall(VecPointwiseMult(shell->left_work, x, shell->left));
+    if (conjugate) { /* get arrays because there is no VecPointwiseMultConj() */
+      PetscInt           i, m;
+      const PetscScalar *d, *xarray;
+      PetscScalar       *w;
+      PetscCall(VecGetLocalSize(x, &m));
+      PetscCall(VecGetArrayRead(shell->left, &d));
+      PetscCall(VecGetArrayRead(x, &xarray));
+      PetscCall(VecGetArrayWrite(shell->left_work, &w));
+      for (i = 0; i < m; i++) w[i] = PetscConj(d[i]) * xarray[i];
+      PetscCall(VecRestoreArrayRead(shell->dshift, &d));
+      PetscCall(VecRestoreArrayRead(x, &xarray));
+      PetscCall(VecRestoreArrayWrite(shell->left_work, &w));
+    } else PetscCall(VecPointwiseMult(shell->left_work, x, shell->left));
     *xx = shell->left_work;
   }
   PetscFunctionReturn(PETSC_SUCCESS);
@@ -192,12 +149,24 @@ static PetscErrorCode MatShellPostScaleLeft(Mat A, Vec x)
 /*
     x = diag(right)*x
 */
-static PetscErrorCode MatShellPostScaleRight(Mat A, Vec x)
+static PetscErrorCode MatShellPostScaleRight(Mat A, Vec x, PetscBool conjugate)
 {
   Mat_Shell *shell = (Mat_Shell *)A->data;
 
   PetscFunctionBegin;
-  if (shell->right) PetscCall(VecPointwiseMult(x, x, shell->right));
+  if (shell->right) {
+    if (conjugate) { /* get arrays because there is no VecPointwiseMultConj() */
+      PetscInt           i, m;
+      const PetscScalar *d;
+      PetscScalar       *xarray;
+      PetscCall(VecGetLocalSize(x, &m));
+      PetscCall(VecGetArrayRead(shell->right, &d));
+      PetscCall(VecGetArray(x, &xarray));
+      for (i = 0; i < m; i++) xarray[i] = PetscConj(d[i]) * xarray[i];
+      PetscCall(VecRestoreArrayRead(shell->dshift, &d));
+      PetscCall(VecRestoreArray(x, &xarray));
+    } else PetscCall(VecPointwiseMult(x, x, shell->right));
+  }
   PetscFunctionReturn(PETSC_SUCCESS);
 }
 
@@ -205,10 +174,14 @@ static PetscErrorCode MatShellPostScaleRight(Mat A, Vec x)
          Y = vscale*Y + diag(dshift)*X + vshift*X
 
          On input Y already contains A*x
+
+         If conjugate=PETSC_TRUE then vscale, dshift, and vshift are conjugated
 */
-static PetscErrorCode MatShellShiftAndScale(Mat A, Vec X, Vec Y)
+static PetscErrorCode MatShellShiftAndScale(Mat A, Vec X, Vec Y, PetscBool conjugate)
 {
-  Mat_Shell *shell = (Mat_Shell *)A->data;
+  Mat_Shell  *shell  = (Mat_Shell *)A->data;
+  PetscScalar vscale = conjugate ? PetscConj(shell->vscale) : shell->vscale;
+  PetscScalar vshift = conjugate ? PetscConj(shell->vshift) : shell->vshift;
 
   PetscFunctionBegin;
   if (shell->dshift) { /* get arrays because there is no VecPointwiseMultAdd() */
@@ -219,14 +192,17 @@ static PetscErrorCode MatShellShiftAndScale(Mat A, Vec X, Vec Y)
     PetscCall(VecGetArrayRead(shell->dshift, &d));
     PetscCall(VecGetArrayRead(X, &x));
     PetscCall(VecGetArray(Y, &y));
-    for (i = 0; i < m; i++) y[i] = shell->vscale * y[i] + d[i] * x[i];
+    if (conjugate)
+      for (i = 0; i < m; i++) y[i] = vscale * y[i] + PetscConj(d[i]) * x[i];
+    else
+      for (i = 0; i < m; i++) y[i] = vscale * y[i] + d[i] * x[i];
     PetscCall(VecRestoreArrayRead(shell->dshift, &d));
     PetscCall(VecRestoreArrayRead(X, &x));
     PetscCall(VecRestoreArray(Y, &y));
   } else {
-    PetscCall(VecScale(Y, shell->vscale));
+    PetscCall(VecScale(Y, vscale));
   }
-  if (shell->vshift != 0.0) PetscCall(VecAXPY(Y, shell->vshift, X)); /* if test is for non-square matrices */
+  if (vshift != 0.0) PetscCall(VecAXPY(Y, vshift, X)); /* if test is for non-square matrices */
   PetscFunctionReturn(PETSC_SUCCESS);
 }
 
@@ -255,7 +231,7 @@ static PetscErrorCode MatShellGetContext_Shell(Mat mat, void *ctx)
 
   Fortran Notes:
   You must write a Fortran interface definition for this
-  function that tells Fortran the Fortran derived data type that you are passing in as the ctx argument.
+  function that tells Fortran the Fortran derived data type that you are passing in as the `ctx` argument.
 
 .seealso: [](ch_matrices), `Mat`, `MATSHELL`, `MatCreateShell()`, `MatShellSetOperation()`, `MatShellSetContext()`
 @*/
@@ -286,7 +262,7 @@ static PetscErrorCode MatZeroRowsColumns_Local_Shell(Mat mat, PetscInt nr, Petsc
   /* Expand/create index set of zeroed rows */
   PetscCall(PetscMalloc1(nr, &idxs));
   for (i = 0; i < nr; i++) idxs[i] = rows[i] + rst;
-  PetscCall(ISCreateGeneral(PETSC_COMM_SELF, nr, idxs, PETSC_OWN_POINTER, &is1));
+  PetscCall(ISCreateGeneral(PetscObjectComm((PetscObject)mat), nr, idxs, PETSC_OWN_POINTER, &is1));
   PetscCall(ISSort(is1));
   PetscCall(VecISSet(shell->zvals, is1, diag));
   if (shell->zrows) {
@@ -323,7 +299,7 @@ static PetscErrorCode MatZeroRowsColumns_Local_Shell(Mat mat, PetscInt nr, Petsc
   if (rc) {
     PetscCall(PetscMalloc1(nc, &idxs));
     for (i = 0; i < nc; i++) idxs[i] = cols[i] + cst;
-    PetscCall(ISCreateGeneral(PETSC_COMM_SELF, nc, idxs, PETSC_OWN_POINTER, &is1));
+    PetscCall(ISCreateGeneral(PetscObjectComm((PetscObject)mat), nc, idxs, PETSC_OWN_POINTER, &is1));
     PetscCall(ISSort(is1));
     if (shell->zcols) {
       PetscCall(ISSum(shell->zcols, is1, &is2));
@@ -495,6 +471,7 @@ static PetscErrorCode MatDestroy_Shell(Mat mat)
   PetscCall(PetscObjectComposeFunction((PetscObject)mat, "MatShellSetContextDestroy_C", NULL));
   PetscCall(PetscObjectComposeFunction((PetscObject)mat, "MatShellSetVecType_C", NULL));
   PetscCall(PetscObjectComposeFunction((PetscObject)mat, "MatShellSetManageScalingShifts_C", NULL));
+  PetscCall(PetscObjectComposeFunction((PetscObject)mat, "MatShellGetScalingShifts_C", NULL));
   PetscCall(PetscObjectComposeFunction((PetscObject)mat, "MatShellSetOperation_C", NULL));
   PetscCall(PetscObjectComposeFunction((PetscObject)mat, "MatShellGetOperation_C", NULL));
   PetscCall(PetscObjectComposeFunction((PetscObject)mat, "MatShellSetMatProductOperation_C", NULL));
@@ -860,11 +837,11 @@ set:
 .vb
   extern PetscErrorCode usersymbolic(Mat, Mat, Mat, void**);
   extern PetscErrorCode usernumeric(Mat, Mat, Mat, void*);
-  extern PetscErrorCode userdestroy(void*);
+  extern PetscErrorCode ctxdestroy(void*);
 
   MatCreateShell(comm, m, n, M, N, ctx, &A);
   MatShellSetMatProductOperation(
-    A, MATPRODUCT_AB, usersymbolic, usernumeric, userdestroy,MATSEQAIJ, MATDENSE
+    A, MATPRODUCT_AB, usersymbolic, usernumeric, ctxdestroy,MATSEQAIJ, MATDENSE
   );
   // create B of type SEQAIJ etc..
   MatProductCreate(A, B, PETSC_NULLPTR, &C);
@@ -896,7 +873,7 @@ PetscErrorCode MatShellSetMatProductOperation(Mat A, MatProductType ptype, Petsc
   PetscCheck(numeric, PetscObjectComm((PetscObject)A), PETSC_ERR_USER, "Missing numeric routine, argument 4");
   PetscAssertPointer(Btype, 6);
   if (Ctype) PetscAssertPointer(Ctype, 7);
-  PetscTryMethod(A, "MatShellSetMatProductOperation_C", (Mat, MatProductType, PetscErrorCode(*)(Mat, Mat, Mat, void **), PetscErrorCode(*)(Mat, Mat, Mat, void *), PetscErrorCode(*)(void *), MatType, MatType), (A, ptype, symbolic, numeric, destroy, Btype, Ctype));
+  PetscTryMethod(A, "MatShellSetMatProductOperation_C", (Mat, MatProductType, PetscErrorCode (*)(Mat, Mat, Mat, void **), PetscErrorCode (*)(Mat, Mat, Mat, void *), PetscErrorCode (*)(void *), MatType, MatType), (A, ptype, symbolic, numeric, destroy, Btype, Ctype));
   PetscFunctionReturn(PETSC_SUCCESS);
 }
 
@@ -997,9 +974,9 @@ static PetscErrorCode MatDuplicate_Shell(Mat mat, MatDuplicateOption op, Mat *M)
   PetscFunctionBegin;
   PetscCall(MatCreateShell(PetscObjectComm((PetscObject)mat), mat->rmap->n, mat->cmap->n, mat->rmap->N, mat->cmap->N, NULL, M));
   ((Mat_Shell *)(*M)->data)->ctxcontainer = ((Mat_Shell *)mat->data)->ctxcontainer;
-  PetscCall(PetscObjectCompose((PetscObject)(*M), "MatShell ctx", (PetscObject)((Mat_Shell *)(*M)->data)->ctxcontainer));
-  PetscCall(PetscObjectChangeTypeName((PetscObject)(*M), ((PetscObject)mat)->type_name));
-  if (op != MAT_DO_NOT_COPY_VALUES) PetscCall(MatCopy(mat, *M, SAME_NONZERO_PATTERN));
+  PetscCall(PetscObjectCompose((PetscObject)*M, "MatShell ctx", (PetscObject)((Mat_Shell *)(*M)->data)->ctxcontainer));
+  PetscCall(PetscObjectChangeTypeName((PetscObject)*M, ((PetscObject)mat)->type_name));
+  if (op == MAT_COPY_VALUES) PetscCall(MatCopy(mat, *M, SAME_NONZERO_PATTERN));
   PetscCall(PetscObjectCopyFortranFunctionPointers((PetscObject)mat, (PetscObject)*M));
   PetscFunctionReturn(PETSC_SUCCESS);
 }
@@ -1020,7 +997,7 @@ static PetscErrorCode MatMult_Shell(Mat A, Vec x, Vec y)
     /* increase the state of the output vector since the user did not update its state themself as should have been done */
     PetscCall(PetscObjectStateIncrease((PetscObject)y));
   }
-  PetscCall(MatShellShiftAndScale(A, xx, y));
+  PetscCall(MatShellShiftAndScale(A, xx, y, PETSC_FALSE));
   PetscCall(MatShellPostScaleLeft(A, y));
   PetscCall(MatShellPostZeroLeft(A, y));
 
@@ -1064,7 +1041,7 @@ static PetscErrorCode MatMultTranspose_Shell(Mat A, Vec x, Vec y)
 
   PetscFunctionBegin;
   PetscCall(MatShellPreZeroLeft(A, x, &xx));
-  PetscCall(MatShellPreScaleLeft(A, xx, &xx));
+  PetscCall(MatShellPreScaleLeft(A, xx, &xx, PETSC_FALSE));
   PetscCall(PetscObjectStateGet((PetscObject)y, &instate));
   PetscCall((*shell->ops->multtranspose)(A, xx, y));
   PetscCall(PetscObjectStateGet((PetscObject)y, &outstate));
@@ -1072,8 +1049,8 @@ static PetscErrorCode MatMultTranspose_Shell(Mat A, Vec x, Vec y)
     /* increase the state of the output vector since the user did not update its state themself as should have been done */
     PetscCall(PetscObjectStateIncrease((PetscObject)y));
   }
-  PetscCall(MatShellShiftAndScale(A, xx, y));
-  PetscCall(MatShellPostScaleRight(A, y));
+  PetscCall(MatShellShiftAndScale(A, xx, y, PETSC_FALSE));
+  PetscCall(MatShellPostScaleRight(A, y, PETSC_FALSE));
   PetscCall(MatShellPostZeroRight(A, y));
 
   if (shell->axpy) {
@@ -1091,6 +1068,41 @@ static PetscErrorCode MatMultTranspose_Shell(Mat A, Vec x, Vec y)
   PetscFunctionReturn(PETSC_SUCCESS);
 }
 
+static PetscErrorCode MatMultHermitianTranspose_Shell(Mat A, Vec x, Vec y)
+{
+  Mat_Shell       *shell = (Mat_Shell *)A->data;
+  Vec              xx;
+  PetscObjectState instate, outstate;
+
+  PetscFunctionBegin;
+  PetscCall(MatShellPreZeroLeft(A, x, &xx));
+  PetscCall(MatShellPreScaleLeft(A, xx, &xx, PETSC_TRUE));
+  PetscCall(PetscObjectStateGet((PetscObject)y, &instate));
+  PetscCall((*shell->ops->multhermitiantranspose)(A, xx, y));
+  PetscCall(PetscObjectStateGet((PetscObject)y, &outstate));
+  if (instate == outstate) {
+    /* increase the state of the output vector since the user did not update its state themself as should have been done */
+    PetscCall(PetscObjectStateIncrease((PetscObject)y));
+  }
+  PetscCall(MatShellShiftAndScale(A, xx, y, PETSC_TRUE));
+  PetscCall(MatShellPostScaleRight(A, y, PETSC_TRUE));
+  PetscCall(MatShellPostZeroRight(A, y));
+
+  if (shell->axpy) {
+    Mat              X;
+    PetscObjectState axpy_state;
+
+    PetscCall(MatShellGetContext(shell->axpy, &X));
+    PetscCall(PetscObjectStateGet((PetscObject)X, &axpy_state));
+    PetscCheck(shell->axpy_state == axpy_state, PetscObjectComm((PetscObject)A), PETSC_ERR_ORDER, "Invalid AXPY state: cannot modify the X matrix passed to MatAXPY(Y,a,X,...)");
+    PetscCall(MatCreateVecs(shell->axpy, shell->axpy_right ? NULL : &shell->axpy_right, shell->axpy_left ? NULL : &shell->axpy_left));
+    PetscCall(VecCopy(x, shell->axpy_left));
+    PetscCall(MatMultHermitianTranspose(shell->axpy, shell->axpy_left, shell->axpy_right));
+    PetscCall(VecAXPY(y, PetscConj(shell->axpy_vscale), shell->axpy_right));
+  }
+  PetscFunctionReturn(PETSC_SUCCESS);
+}
+
 static PetscErrorCode MatMultTransposeAdd_Shell(Mat A, Vec x, Vec y, Vec z)
 {
   Mat_Shell *shell = (Mat_Shell *)A->data;
@@ -1107,6 +1119,22 @@ static PetscErrorCode MatMultTransposeAdd_Shell(Mat A, Vec x, Vec y, Vec z)
   PetscFunctionReturn(PETSC_SUCCESS);
 }
 
+static PetscErrorCode MatMultHermitianTransposeAdd_Shell(Mat A, Vec x, Vec y, Vec z)
+{
+  Mat_Shell *shell = (Mat_Shell *)A->data;
+
+  PetscFunctionBegin;
+  if (y == z) {
+    if (!shell->left_add_work) PetscCall(VecDuplicate(z, &shell->left_add_work));
+    PetscCall(MatMultHermitianTranspose(A, x, shell->left_add_work));
+    PetscCall(VecAXPY(z, 1.0, shell->left_add_work));
+  } else {
+    PetscCall(MatMultHermitianTranspose(A, x, z));
+    PetscCall(VecAXPY(z, 1.0, y));
+  }
+  PetscFunctionReturn(PETSC_SUCCESS);
+}
+
 /*
           diag(left)(vscale*A + diag(dshift) + vshift I)diag(right)
 */
@@ -1141,6 +1169,40 @@ static PetscErrorCode MatGetDiagonal_Shell(Mat A, Vec v)
   PetscFunctionReturn(PETSC_SUCCESS);
 }
 
+static PetscErrorCode MatGetDiagonalBlock_Shell(Mat A, Mat *b)
+{
+  Mat_Shell *shell = (Mat_Shell *)A->data;
+  Vec        left = NULL, right = NULL;
+
+  PetscFunctionBegin;
+  PetscCheck(!shell->zrows && !shell->zcols, PetscObjectComm((PetscObject)A), PETSC_ERR_SUP, "Cannot call MatGetDiagonalBlock() if MatZeroRows() or MatZeroRowsColumns() has been called on the input Mat"); // TODO FIXME
+  PetscCheck(!shell->axpy, PetscObjectComm((PetscObject)A), PETSC_ERR_SUP, "Cannot call MatGetDiagonalBlock() if MatAXPY() has been called on the input Mat");                                               // TODO FIXME
+  PetscCheck(!shell->dshift, PetscObjectComm((PetscObject)A), PETSC_ERR_SUP, "Cannot call MatGetDiagonalBlock() if MatDiagonalSet() has been called on the input Mat");                                      // TODO FIXME
+  if (shell->ops->getdiagonalblock) {
+    PetscCall((*shell->ops->getdiagonalblock)(A, b));
+  } else SETERRQ(PetscObjectComm((PetscObject)A), PETSC_ERR_ARG_WRONGSTATE, "Must provide shell matrix with routine to return diagonal block using\nMatShellSetOperation(S,MATOP_GET_DIAGONAL_BLOCK,...)");
+  PetscCall(MatScale(*b, shell->vscale));
+  PetscCall(MatShift(*b, shell->vshift));
+  if (shell->left) {
+    PetscCall(VecCreateLocalVector(shell->left, &left));
+    PetscCall(VecGetLocalVectorRead(shell->left, left));
+  }
+  if (shell->right) {
+    PetscCall(VecCreateLocalVector(shell->right, &right));
+    PetscCall(VecGetLocalVectorRead(shell->right, right));
+  }
+  PetscCall(MatDiagonalScale(*b, left, right));
+  if (shell->left) {
+    PetscCall(VecRestoreLocalVectorRead(shell->left, left));
+    PetscCall(VecDestroy(&left));
+  }
+  if (shell->right) {
+    PetscCall(VecRestoreLocalVectorRead(shell->right, right));
+    PetscCall(VecDestroy(&right));
+  }
+  PetscFunctionReturn(PETSC_SUCCESS);
+}
+
 static PetscErrorCode MatShift_Shell(Mat Y, PetscScalar a)
 {
   Mat_Shell *shell = (Mat_Shell *)Y->data;
@@ -1257,7 +1319,7 @@ static PetscErrorCode MatDiagonalScale_Shell(Mat Y, Vec left, Vec right)
   PetscFunctionReturn(PETSC_SUCCESS);
 }
 
-static PetscErrorCode MatAssemblyEnd_Shell(Mat Y, MatAssemblyType t)
+PETSC_INTERN PetscErrorCode MatAssemblyEnd_Shell(Mat Y, MatAssemblyType t)
 {
   Mat_Shell *shell = (Mat_Shell *)Y->data;
 
@@ -1339,7 +1401,7 @@ static struct _MatOps MatOps_Values = {NULL,
                                        /*29*/ NULL,
                                        NULL,
                                        NULL,
-                                       NULL,
+                                       /*32*/ NULL,
                                        NULL,
                                        /*34*/ MatDuplicate_Shell,
                                        NULL,
@@ -1429,7 +1491,7 @@ static struct _MatOps MatOps_Values = {NULL,
                                        /*119*/ NULL,
                                        NULL,
                                        NULL,
-                                       NULL,
+                                       MatMultHermitianTransposeAdd_Shell,
                                        NULL,
                                        /*124*/ NULL,
                                        NULL,
@@ -1458,6 +1520,11 @@ static struct _MatOps MatOps_Values = {NULL,
                                        NULL,
                                        NULL,
                                        /*150*/ NULL,
+                                       NULL,
+                                       NULL,
+                                       NULL,
+                                       NULL,
+                                       /*155*/ NULL,
                                        NULL};
 
 static PetscErrorCode MatShellSetContext_Shell(Mat mat, void *ctx)
@@ -1476,16 +1543,36 @@ static PetscErrorCode MatShellSetContext_Shell(Mat mat, void *ctx)
     PetscCall(PetscObjectCompose((PetscObject)mat, "MatShell ctx", NULL));
     shell->ctxcontainer = NULL;
   }
-
   PetscFunctionReturn(PETSC_SUCCESS);
 }
 
-static PetscErrorCode MatShellSetContextDestroy_Shell(Mat mat, PetscErrorCode (*f)(void *))
+static PetscErrorCode MatShellSetContextDestroy_Shell(Mat mat, PetscCtxDestroyFn *f)
 {
   Mat_Shell *shell = (Mat_Shell *)mat->data;
 
   PetscFunctionBegin;
-  if (shell->ctxcontainer) PetscCall(PetscContainerSetUserDestroy(shell->ctxcontainer, f));
+  if (shell->ctxcontainer) PetscCall(PetscContainerSetCtxDestroy(shell->ctxcontainer, f));
+  PetscFunctionReturn(PETSC_SUCCESS);
+}
+
+PetscErrorCode MatShellSetContext_Immutable(Mat mat, void *ctx)
+{
+  PetscFunctionBegin;
+  SETERRQ(PetscObjectComm((PetscObject)mat), PETSC_ERR_ARG_WRONGSTATE, "Cannot call MatShellSetContext() for a %s, it is used internally by the structure", ((PetscObject)mat)->type_name);
+  PetscFunctionReturn(PETSC_SUCCESS);
+}
+
+PetscErrorCode MatShellSetContextDestroy_Immutable(Mat mat, PetscCtxDestroyFn *f)
+{
+  PetscFunctionBegin;
+  SETERRQ(PetscObjectComm((PetscObject)mat), PETSC_ERR_ARG_WRONGSTATE, "Cannot call MatShellSetContextDestroy() for a %s, it is used internally by the structure", ((PetscObject)mat)->type_name);
+  PetscFunctionReturn(PETSC_SUCCESS);
+}
+
+PetscErrorCode MatShellSetManageScalingShifts_Immutable(Mat mat)
+{
+  PetscFunctionBegin;
+  SETERRQ(PetscObjectComm((PetscObject)mat), PETSC_ERR_ARG_WRONGSTATE, "Cannot call MatShellSetManageScalingShifts() for a %s, it is used internally by the structure", ((PetscObject)mat)->type_name);
   PetscFunctionReturn(PETSC_SUCCESS);
 }
 
@@ -1511,6 +1598,31 @@ static PetscErrorCode MatShellSetManageScalingShifts_Shell(Mat A)
   PetscFunctionReturn(PETSC_SUCCESS);
 }
 
+static PetscErrorCode MatShellGetScalingShifts_Shell(Mat A, PetscScalar *vshift, PetscScalar *vscale, Vec *dshift, Vec *left, Vec *right, Mat *axpy, IS *zrows, IS *zcols)
+{
+  Mat_Shell *shell = (Mat_Shell *)A->data;
+
+  PetscFunctionBegin;
+  PetscCheck(shell->managescalingshifts, PetscObjectComm((PetscObject)A), PETSC_ERR_SUP, "Cannot get scaling and shifts if MatShellSetManageScalingShifts() has been called");
+  if (vshift == MAT_SHELL_NOT_ALLOWED) PetscCheck(shell->vshift == 0.0, PetscObjectComm((PetscObject)A), PETSC_ERR_SUP, "Non-trivial member in parent MatShell: vshift != 0.0, set via MatShift()");
+  else if (vshift) *vshift = shell->vshift;
+  if (vscale == MAT_SHELL_NOT_ALLOWED) PetscCheck(shell->vscale == 1.0, PetscObjectComm((PetscObject)A), PETSC_ERR_SUP, "Non-trivial member in parent MatShell: vscale != 1.0, set via MatScale()");
+  else if (vscale) *vscale = shell->vscale;
+  if (dshift == MAT_SHELL_NOT_ALLOWED) PetscCheck(!shell->dshift, PetscObjectComm((PetscObject)A), PETSC_ERR_SUP, "Non-trivial member in parent MatShell: dshift, set via MatDiagonalSet()");
+  else if (dshift) *dshift = shell->dshift;
+  if (left == MAT_SHELL_NOT_ALLOWED) PetscCheck(!shell->left, PetscObjectComm((PetscObject)A), PETSC_ERR_SUP, "Non-trivial member in parent MatShell: left, set via MatDiagonalScale()");
+  else if (left) *left = shell->left;
+  if (right == MAT_SHELL_NOT_ALLOWED) PetscCheck(!shell->right, PetscObjectComm((PetscObject)A), PETSC_ERR_SUP, "Non-trivial member in parent MatShell: right, set via MatDiagonalScale()");
+  else if (right) *right = shell->right;
+  if (axpy == MAT_SHELL_NOT_ALLOWED) PetscCheck(!shell->axpy, PetscObjectComm((PetscObject)A), PETSC_ERR_SUP, "Non-trivial member in parent MatShell: axpy, set via MatAXPY()");
+  else if (axpy) *axpy = shell->axpy;
+  if (zrows == MAT_SHELL_NOT_ALLOWED) PetscCheck(!shell->zrows, PetscObjectComm((PetscObject)A), PETSC_ERR_SUP, "Non-trivial member in parent MatShell: zrows, set via MatZeroRows()");
+  else if (zrows) *zrows = shell->zrows;
+  if (zcols == MAT_SHELL_NOT_ALLOWED) PetscCheck(!shell->zcols, PetscObjectComm((PetscObject)A), PETSC_ERR_SUP, "Non-trivial member in parent MatShell: zcols, set via MatZeroRowsColumns()");
+  else if (zcols) *zcols = shell->zcols;
+  PetscFunctionReturn(PETSC_SUCCESS);
+}
+
 static PetscErrorCode MatShellSetOperation_Shell(Mat mat, MatOperation op, void (*f)(void))
 {
   Mat_Shell *shell = (Mat_Shell *)mat->data;
@@ -1518,14 +1630,14 @@ static PetscErrorCode MatShellSetOperation_Shell(Mat mat, MatOperation op, void
   PetscFunctionBegin;
   switch (op) {
   case MATOP_DESTROY:
-    shell->ops->destroy = (PetscErrorCode(*)(Mat))f;
+    shell->ops->destroy = (PetscErrorCode (*)(Mat))f;
     break;
   case MATOP_VIEW:
     if (!mat->ops->viewnative) mat->ops->viewnative = mat->ops->view;
-    mat->ops->view = (PetscErrorCode(*)(Mat, PetscViewer))f;
+    mat->ops->view = (PetscErrorCode (*)(Mat, PetscViewer))f;
     break;
   case MATOP_COPY:
-    shell->ops->copy = (PetscErrorCode(*)(Mat, Mat, MatStructure))f;
+    shell->ops->copy = (PetscErrorCode (*)(Mat, Mat, MatStructure))f;
     break;
   case MATOP_DIAGONAL_SET:
   case MATOP_DIAGONAL_SCALE:
@@ -1539,29 +1651,47 @@ static PetscErrorCode MatShellSetOperation_Shell(Mat mat, MatOperation op, void
     break;
   case MATOP_GET_DIAGONAL:
     if (shell->managescalingshifts) {
-      shell->ops->getdiagonal = (PetscErrorCode(*)(Mat, Vec))f;
+      shell->ops->getdiagonal = (PetscErrorCode (*)(Mat, Vec))f;
       mat->ops->getdiagonal   = MatGetDiagonal_Shell;
     } else {
       shell->ops->getdiagonal = NULL;
-      mat->ops->getdiagonal   = (PetscErrorCode(*)(Mat, Vec))f;
+      mat->ops->getdiagonal   = (PetscErrorCode (*)(Mat, Vec))f;
+    }
+    break;
+  case MATOP_GET_DIAGONAL_BLOCK:
+    if (shell->managescalingshifts) {
+      shell->ops->getdiagonalblock = (PetscErrorCode (*)(Mat, Mat *))f;
+      mat->ops->getdiagonalblock   = MatGetDiagonalBlock_Shell;
+    } else {
+      shell->ops->getdiagonalblock = NULL;
+      mat->ops->getdiagonalblock   = (PetscErrorCode (*)(Mat, Mat *))f;
     }
     break;
   case MATOP_MULT:
     if (shell->managescalingshifts) {
-      shell->ops->mult = (PetscErrorCode(*)(Mat, Vec, Vec))f;
+      shell->ops->mult = (PetscErrorCode (*)(Mat, Vec, Vec))f;
       mat->ops->mult   = MatMult_Shell;
     } else {
       shell->ops->mult = NULL;
-      mat->ops->mult   = (PetscErrorCode(*)(Mat, Vec, Vec))f;
+      mat->ops->mult   = (PetscErrorCode (*)(Mat, Vec, Vec))f;
     }
     break;
   case MATOP_MULT_TRANSPOSE:
     if (shell->managescalingshifts) {
-      shell->ops->multtranspose = (PetscErrorCode(*)(Mat, Vec, Vec))f;
+      shell->ops->multtranspose = (PetscErrorCode (*)(Mat, Vec, Vec))f;
       mat->ops->multtranspose   = MatMultTranspose_Shell;
     } else {
       shell->ops->multtranspose = NULL;
-      mat->ops->multtranspose   = (PetscErrorCode(*)(Mat, Vec, Vec))f;
+      mat->ops->multtranspose   = (PetscErrorCode (*)(Mat, Vec, Vec))f;
+    }
+    break;
+  case MATOP_MULT_HERMITIAN_TRANSPOSE:
+    if (shell->managescalingshifts) {
+      shell->ops->multhermitiantranspose = (PetscErrorCode (*)(Mat, Vec, Vec))f;
+      mat->ops->multhermitiantranspose   = MatMultHermitianTranspose_Shell;
+    } else {
+      shell->ops->multhermitiantranspose = NULL;
+      mat->ops->multhermitiantranspose   = (PetscErrorCode (*)(Mat, Vec, Vec))f;
     }
     break;
   default:
@@ -1599,6 +1729,10 @@ static PetscErrorCode MatShellGetOperation_Shell(Mat mat, MatOperation op, void
     if (shell->ops->getdiagonal) *f = (void (*)(void))shell->ops->getdiagonal;
     else *f = (((void (**)(void))mat->ops)[op]);
     break;
+  case MATOP_GET_DIAGONAL_BLOCK:
+    if (shell->ops->getdiagonalblock) *f = (void (*)(void))shell->ops->getdiagonalblock;
+    else *f = (((void (**)(void))mat->ops)[op]);
+    break;
   case MATOP_MULT:
     if (shell->ops->mult) *f = (void (*)(void))shell->ops->mult;
     else *f = (((void (**)(void))mat->ops)[op]);
@@ -1607,6 +1741,10 @@ static PetscErrorCode MatShellGetOperation_Shell(Mat mat, MatOperation op, void
     if (shell->ops->multtranspose) *f = (void (*)(void))shell->ops->multtranspose;
     else *f = (((void (**)(void))mat->ops)[op]);
     break;
+  case MATOP_MULT_HERMITIAN_TRANSPOSE:
+    if (shell->ops->multhermitiantranspose) *f = (void (*)(void))shell->ops->multhermitiantranspose;
+    else *f = (((void (**)(void))mat->ops)[op]);
+    break;
   default:
     *f = (((void (**)(void))mat->ops)[op]);
   }
@@ -1614,11 +1752,18 @@ static PetscErrorCode MatShellGetOperation_Shell(Mat mat, MatOperation op, void
 }
 
 /*MC
-   MATSHELL - MATSHELL = "shell" - A matrix type to be used to define your own matrix type -- perhaps matrix-free.
+  MATSHELL - MATSHELL = "shell" - A matrix type to be used to define your own matrix type with its own data structure -- perhaps matrix-free.
 
   Level: advanced
 
-.seealso: [](ch_matrices), `Mat`, `MatCreateShell()`
+  Notes:
+  See `MatCreateShell()` for details on the usage of `MATSHELL`
+
+  `PCSHELL` can be used in conjunction with `MATSHELL` to provide a custom preconditioner appropriate for your `MATSHELL`. Since
+  many standard preconditioners such as `PCILU` depend on having an explicit representation of the matrix entries they cannot be used
+  directly with `MATSHELL`.
+
+.seealso: [](ch_matrices), `Mat`, `MatCreateShell()`, `PCSHELL`
 M*/
 
 PETSC_EXTERN PetscErrorCode MatCreate_Shell(Mat A)
@@ -1642,6 +1787,7 @@ PETSC_EXTERN PetscErrorCode MatCreate_Shell(Mat A)
   PetscCall(PetscObjectComposeFunction((PetscObject)A, "MatShellSetContextDestroy_C", MatShellSetContextDestroy_Shell));
   PetscCall(PetscObjectComposeFunction((PetscObject)A, "MatShellSetVecType_C", MatShellSetVecType_Shell));
   PetscCall(PetscObjectComposeFunction((PetscObject)A, "MatShellSetManageScalingShifts_C", MatShellSetManageScalingShifts_Shell));
+  PetscCall(PetscObjectComposeFunction((PetscObject)A, "MatShellGetScalingShifts_C", MatShellGetScalingShifts_Shell));
   PetscCall(PetscObjectComposeFunction((PetscObject)A, "MatShellSetOperation_C", MatShellSetOperation_Shell));
   PetscCall(PetscObjectComposeFunction((PetscObject)A, "MatShellGetOperation_C", MatShellGetOperation_Shell));
   PetscCall(PetscObjectComposeFunction((PetscObject)A, "MatShellSetMatProductOperation_C", MatShellSetMatProductOperation_Shell));
@@ -1651,7 +1797,7 @@ PETSC_EXTERN PetscErrorCode MatCreate_Shell(Mat A)
 
 /*@C
   MatCreateShell - Creates a new matrix of `MatType` `MATSHELL` for use with a user-defined
-  private data storage format.
+  private matrix data storage format.
 
   Collective
 
@@ -1674,14 +1820,15 @@ PETSC_EXTERN PetscErrorCode MatCreate_Shell(Mat A)
 
   MatCreateShell(comm, m, n, M, N, ctx, &mat);
   MatShellSetOperation(mat, MATOP_MULT, (void(*)(void))mult);
+  MatShellSetContext(mat,ctx);
   // Use matrix for operations that have been set
   MatDestroy(mat);
 .ve
 
   Notes:
-  The shell matrix type is intended to provide a simple class to use
-  with `KSP` (such as, for use with matrix-free methods). You should not
-  use the shell type if you plan to define a complete matrix class.
+  The shell matrix type is intended to provide a simple way for users to write a custom matrix specifically for their application.
+
+  `MatCreateShell()` is used in conjunction with `MatShellSetContext()` and `MatShellSetOperation()`.
 
   PETSc requires that matrices and vectors being used for certain
   operations are partitioned accordingly.  For example, when
@@ -1731,7 +1878,7 @@ PETSC_EXTERN PetscErrorCode MatCreate_Shell(Mat A)
   an update in the preconditioner you must call `MatAssemblyBegin()` and `MatAssemblyEnd()` or `PetscObjectStateIncrease`((`PetscObject`)mat);
   each time the `MATSHELL` matrix has changed.
 
-  Matrix product operations (i.e. `MatMat()`, `MatTransposeMat()` etc) can be specified using `MatShellSetMatProductOperation()`
+  Matrix-matrix product operations can be specified using `MatShellSetMatProductOperation()`
 
   Calling `MatAssemblyBegin()`/`MatAssemblyEnd()` on a `MATSHELL` removes any previously supplied shift and scales that were provided
   with `MatDiagonalSet()`, `MatShift()`, `MatScale()`, or `MatDiagonalScale()`.
@@ -1765,6 +1912,10 @@ PetscErrorCode MatCreateShell(MPI_Comm comm, PetscInt m, PetscInt n, PetscInt M,
 
   Level: advanced
 
+  Note:
+  This provides an easy way, along with `MatCreateShell()` and `MatShellSetOperation()` to provide a custom matrix format
+  specifically for your application.
+
   Fortran Notes:
   You must write a Fortran interface definition for this
   function that tells Fortran the Fortran derived data type that you are passing in as the `ctx` argument.
@@ -1786,7 +1937,7 @@ PetscErrorCode MatShellSetContext(Mat mat, void *ctx)
 
   Input Parameters:
 + mat - the shell matrix
-- f   - the context destroy function
+- f   - the context destroy function, see `PetscCtxDestroyFn` for calling sequence
 
   Level: advanced
 
@@ -1796,13 +1947,14 @@ PetscErrorCode MatShellSetContext(Mat mat, void *ctx)
   ensures proper reference counting for the user provided context data in the case that
   the `MATSHELL` is duplicated.
 
-.seealso: [](ch_matrices), `Mat`, `MATSHELL`, `MatCreateShell()`, `MatShellSetContext()`
+.seealso: [](ch_matrices), `Mat`, `MATSHELL`, `MatCreateShell()`, `MatShellSetContext()`,
+          `PetscCtxDestroyFn`
 @*/
-PetscErrorCode MatShellSetContextDestroy(Mat mat, PetscErrorCode (*f)(void *))
+PetscErrorCode MatShellSetContextDestroy(Mat mat, PetscCtxDestroyFn *f)
 {
   PetscFunctionBegin;
   PetscValidHeaderSpecific(mat, MAT_CLASSID, 1);
-  PetscTryMethod(mat, "MatShellSetContextDestroy_C", (Mat, PetscErrorCode(*)(void *)), (mat, f));
+  PetscTryMethod(mat, "MatShellSetContextDestroy_C", (Mat, PetscCtxDestroyFn *), (mat, f));
   PetscFunctionReturn(PETSC_SUCCESS);
 }
 
@@ -1847,6 +1999,42 @@ PetscErrorCode MatShellSetManageScalingShifts(Mat A)
   PetscFunctionReturn(PETSC_SUCCESS);
 }
 
+/*@C
+  MatShellGetScalingShifts - Gets members of a `MATSHELL` used internally for scaling and
+  shifting the `Mat` or calling `MatAXPY()`, `MatZeroRows()`, or `MatZeroRowsColumns()` with it
+
+  Logically Collective
+
+  Input Parameter:
+. A - the `MATSHELL` shell matrix
+
+  Output Parameters:
++ vshift - `PetscScalar` pointer (can be `NULL`), see `MatShift()`, or `MAT_SHELL_NOT_ALLOWED` if the internal shift should be 0
+. vscale - `PetscScalar` pointer (can be `NULL`), see `MatScale()`, or `MAT_SHELL_NOT_ALLOWED` if the internal scaling should be 1
+. dshift - `Vec` pointer (can be `NULL`), see `MatDiagonalSet()`, or `MAT_SHELL_NOT_ALLOWED` if the internal shift should be `NULL`
+. left   - `Vec` pointer (can be `NULL`), see `MatDiagonalScale()`, or `MAT_SHELL_NOT_ALLOWED` if the internal scaling should be `NULL`
+. right  - `Vec` pointer (can be `NULL`), see `MatDiagonalScale()`, or `MAT_SHELL_NOT_ALLOWED` if the internal scaling should be `NULL`
+. axpy   - `Mat` pointer (can be `NULL`), or `MAT_SHELL_NOT_ALLOWED` if `MatAXPY()` should have not been called on `A`
+. zrows  - `Vec` pointer (can be `NULL`), or `MAT_SHELL_NOT_ALLOWED` if `MatZeroRows()` should have not been called on `A`
+- zcols  - `Vec` pointer (can be `NULL`), or `MAT_SHELL_NOT_ALLOWED` if `MatZeroRowsColumns()` should have not been called on `A`
+
+  Level: advanced
+
+  Developer Notes:
+  This is mostly useful to check for corner-cases in `MatType` deriving from
+  `MATSHELL`, e.g, `MATCOMPOSITE` or `MATTRANSPOSEVIRTUAL`, since scaling and
+  shifts often require extra work which is not always implemented.
+
+.seealso: [](ch_matrices), `Mat`, `MATSHELL`, `MatCreateShell()`, `MatShift()`, `MatScale()`, `MatDiagonalSet()`, `MatDiagonalScale()`, `MatAXPY()`, `MatZeroRows()`, `MatZeroRowsColumns()`, `MatShellSetManageScalingShifts()`
+@*/
+PetscErrorCode MatShellGetScalingShifts(Mat A, PetscScalar *vshift, PetscScalar *vscale, Vec *dshift, Vec *left, Vec *right, Mat *axpy, IS *zrows, IS *zcols)
+{
+  PetscFunctionBegin;
+  PetscValidHeaderSpecific(A, MAT_CLASSID, 1);
+  PetscTryMethod(A, "MatShellGetScalingShifts_C", (Mat, PetscScalar *, PetscScalar *, Vec *, Vec *, Vec *, Mat *, IS *, IS *), (A, vshift, vscale, dshift, left, right, axpy, zrows, zcols));
+  PetscFunctionReturn(PETSC_SUCCESS);
+}
+
 /*@C
   MatShellTestMult - Compares the multiply routine provided to the `MATSHELL` with differencing on a given function.
 
@@ -2008,7 +2196,9 @@ PetscErrorCode MatShellTestMultTranspose(Mat mat, PetscErrorCode (*f)(void *, Ve
   sequence as the usual matrix interface routines, since they
   are intended to be accessed via the usual matrix interface
   routines, e.g.,
-$       MatMult(Mat, Vec, Vec) -> usermult(Mat, Vec, Vec)
+.vb
+  MatMult(Mat, Vec, Vec) -> usermult(Mat, Vec, Vec)
+.ve
 
   In particular each function MUST return an error code of 0 on success and
   nonzero on failure.
@@ -2020,7 +2210,7 @@ $       MatMult(Mat, Vec, Vec) -> usermult(Mat, Vec, Vec)
   Use `MatSetOperation()` to set an operation for any matrix type. For matrix product operations (i.e. `MatMatXXX()`, `MatTransposeMatXXX()` etc)
   use `MatShellSetMatProductOperation()`
 
-  Fortran Notes:
+  Fortran Note:
   For `MatCreateVecs()` the user code should check if the input left or right matrix is -1 and in that case not
   generate a matrix. See src/mat/tests/ex120f.F
 
@@ -2058,7 +2248,9 @@ PetscErrorCode MatShellSetOperation(Mat mat, MatOperation op, void (*g)(void))
   sequence as the usual matrix interface routines, since they
   are intended to be accessed via the usual matrix interface
   routines, e.g.,
-$       MatMult(Mat, Vec, Vec) -> usermult(Mat, Vec, Vec)
+.vb
+  MatMult(Mat, Vec, Vec) -> usermult(Mat, Vec, Vec)
+.ve
 
   Within each user-defined routine, the user should call
   `MatShellGetContext()` to obtain the user-defined context that was
@@ -2081,14 +2273,10 @@ PetscErrorCode MatShellGetOperation(Mat mat, MatOperation op, void (**g)(void))
 . mat - the matrix
 
   Output Parameter:
-. flg - the boolean value
+. flg - the Boolean value
 
   Level: developer
 
-  Developer Notes:
-  In the future, we should allow the object type name to be changed still using the `MATSHELL` data structure for other matrices
-  (i.e. `MATTRANSPOSEVIRTUAL`, `MATSCHURCOMPLEMENT` etc)
-
 .seealso: [](ch_matrices), `Mat`, `MATSHELL`, `MATMFFD`, `MatCreateShell()`, `MATTRANSPOSEVIRTUAL`, `MATSCHURCOMPLEMENT`
 @*/
 PetscErrorCode MatIsShell(Mat mat, PetscBool *flg)
