@@ -53,12 +53,12 @@ static PetscErrorCode TaoADMMToleranceUpdate(Tao tao)
    * Set gatol to be gatol_admm *  ||A^Ty|| *
    * while cnorm is ||r_k||_2, and gnorm is ||d_k||_2 */
   temp = am->catol_admm * PetscMax(Axnorm, (!am->const_norm) ? Bznorm : PetscMax(Bznorm, am->const_norm));
-  PetscCall(TaoSetConstraintTolerances(tao, temp, PETSC_DEFAULT));
-  PetscCall(TaoSetTolerances(tao, am->gatol_admm * ATynorm, PETSC_DEFAULT, PETSC_DEFAULT));
+  PetscCall(TaoSetConstraintTolerances(tao, temp, PETSC_CURRENT));
+  PetscCall(TaoSetTolerances(tao, am->gatol_admm * ATynorm, PETSC_CURRENT, PETSC_CURRENT));
   PetscFunctionReturn(PETSC_SUCCESS);
 }
 
-/* Penaly Update for Adaptive ADMM. */
+/* Penalty Update for Adaptive ADMM. */
 static PetscErrorCode AdaptiveADMMPenaltyUpdate(Tao tao)
 {
   TAO_ADMM *am = (TAO_ADMM *)tao->data;
@@ -354,11 +354,11 @@ static PetscErrorCode TaoSolve_ADMM(Tao tao)
 
   if (!am->zJI) {
     /* Currently, B is assumed to be a linear system, i.e., not getting updated*/
-    PetscCall(MatTransposeMatMult(am->JB, am->JB, MAT_INITIAL_MATRIX, PETSC_DEFAULT, &am->BTB));
+    PetscCall(MatTransposeMatMult(am->JB, am->JB, MAT_INITIAL_MATRIX, PETSC_DETERMINE, &am->BTB));
   }
   if (!am->xJI) {
     /* Currently, A is assumed to be a linear system, i.e., not getting updated*/
-    PetscCall(MatTransposeMatMult(am->subsolverX->jacobian_equality, am->subsolverX->jacobian_equality, MAT_INITIAL_MATRIX, PETSC_DEFAULT, &am->ATA));
+    PetscCall(MatTransposeMatMult(am->subsolverX->jacobian_equality, am->subsolverX->jacobian_equality, MAT_INITIAL_MATRIX, PETSC_DETERMINE, &am->ATA));
   }
 
   is_reg_shell = PETSC_FALSE;
@@ -502,7 +502,7 @@ static PetscErrorCode TaoSolve_ADMM(Tao tao)
   PetscFunctionReturn(PETSC_SUCCESS);
 }
 
-static PetscErrorCode TaoSetFromOptions_ADMM(Tao tao, PetscOptionItems *PetscOptionsObject)
+static PetscErrorCode TaoSetFromOptions_ADMM(Tao tao, PetscOptionItems PetscOptionsObject)
 {
   TAO_ADMM *am = (TAO_ADMM *)tao->data;
 
@@ -604,8 +604,8 @@ static PetscErrorCode TaoSetUp_ADMM(Tao tao)
   }
 
   /* Save changed tao tolerance for adaptive tolerance */
-  if (tao->gatol_changed) am->gatol_admm = tao->gatol;
-  if (tao->catol_changed) am->catol_admm = tao->catol;
+  if (tao->gatol != tao->default_gatol) am->gatol_admm = tao->gatol;
+  if (tao->catol != tao->default_catol) am->catol_admm = tao->catol;
 
   /*Update spectral and dual elements to X subsolver */
   PetscCall(TaoSetObjectiveAndGradient(am->subsolverX, NULL, SubObjGradUpdate, tao));
@@ -709,6 +709,8 @@ PETSC_EXTERN PetscErrorCode TaoCreate_ADMM(Tao tao)
   tao->ops->view           = TaoView_ADMM;
   tao->ops->solve          = TaoSolve_ADMM;
 
+  PetscCall(TaoParametersInitialize(tao));
+
   tao->data           = (void *)am;
   am->l1epsilon       = 1e-6;
   am->lambda          = 1e-4;
