@@ -1,16 +1,15 @@
 #include <../src/vec/is/sf/impls/basic/sfpack.h>
 
 #include <petsc_kokkos.hpp>
+#include <petsc/private/kokkosimpl.hpp>
 
 using DeviceExecutionSpace = Kokkos::DefaultExecutionSpace;
-using DeviceMemorySpace    = typename DeviceExecutionSpace::memory_space;
-using HostMemorySpace      = Kokkos::HostSpace;
 
-typedef Kokkos::View<char *, DeviceMemorySpace> deviceBuffer_t;
-typedef Kokkos::View<char *, HostMemorySpace>   HostBuffer_t;
+typedef Kokkos::View<char *, DefaultMemorySpace>    deviceBuffer_t;
+typedef Kokkos::View<char *, HostMirrorMemorySpace> HostBuffer_t;
 
-typedef Kokkos::View<const char *, DeviceMemorySpace> deviceConstBuffer_t;
-typedef Kokkos::View<const char *, HostMemorySpace>   HostConstBuffer_t;
+typedef Kokkos::View<const char *, DefaultMemorySpace>    deviceConstBuffer_t;
+typedef Kokkos::View<const char *, HostMirrorMemorySpace> HostConstBuffer_t;
 
 /*====================================================================================*/
 /*                             Regular operations                           */
@@ -140,7 +139,7 @@ struct Maxloc {
 /*====================================================================================*/
 template <typename Type>
 struct AtomicInsert {
-  KOKKOS_INLINE_FUNCTION void operator()(Type &x, Type y) const { Kokkos::atomic_assign(&x, y); }
+  KOKKOS_INLINE_FUNCTION void operator()(Type &x, Type y) const { Kokkos::atomic_store(&x, y); }
 };
 template <typename Type>
 struct AtomicAdd {
@@ -213,7 +212,7 @@ static KOKKOS_INLINE_FUNCTION PetscInt MapTidToIndex(const PetscInt *opt, PetscI
   j = (m - k * dx[r] * dy[r]) / dx[r];
   i = m - k * dx[r] * dy[r] - j * dx[r];
 
-  return (start[r] + k * X[r] * Y[r] + j * X[r] + i);
+  return start[r] + k * X[r] * Y[r] + j * X[r] + i;
 }
 
 /*====================================================================================*/
@@ -232,11 +231,11 @@ static KOKKOS_INLINE_FUNCTION PetscInt MapTidToIndex(const PetscInt *opt, PetscI
 template <typename Type, PetscInt BS, PetscInt EQ>
 static PetscErrorCode Pack(PetscSFLink link, PetscInt count, PetscInt start, PetscSFPackOpt opt, const PetscInt *idx, const void *data_, void *buf_)
 {
-  const PetscInt       *iopt = opt ? opt->array : NULL;
-  const PetscInt        M = EQ ? 1 : link->bs / BS, MBS = M * BS; /* If EQ, then MBS will be a compile-time const */
-  const Type           *data = static_cast<const Type *>(data_);
-  Type                 *buf  = static_cast<Type *>(buf_);
-  DeviceExecutionSpace &exec = PetscGetKokkosExecutionSpace();
+  const PetscInt      *iopt = opt ? opt->array : NULL;
+  const PetscInt       M = EQ ? 1 : link->bs / BS, MBS = M * BS; /* If EQ, then MBS will be a compile-time const */
+  const Type          *data = static_cast<const Type *>(data_);
+  Type                *buf  = static_cast<Type *>(buf_);
+  DeviceExecutionSpace exec = PetscGetKokkosExecutionSpace();
 
   PetscFunctionBegin;
   Kokkos::parallel_for(
@@ -254,12 +253,12 @@ static PetscErrorCode Pack(PetscSFLink link, PetscInt count, PetscInt start, Pet
 template <typename Type, class Op, PetscInt BS, PetscInt EQ>
 static PetscErrorCode UnpackAndOp(PetscSFLink link, PetscInt count, PetscInt start, PetscSFPackOpt opt, const PetscInt *idx, void *data_, const void *buf_)
 {
-  Op                    op;
-  const PetscInt       *iopt = opt ? opt->array : NULL;
-  const PetscInt        M = EQ ? 1 : link->bs / BS, MBS = M * BS;
-  Type                 *data = static_cast<Type *>(data_);
-  const Type           *buf  = static_cast<const Type *>(buf_);
-  DeviceExecutionSpace &exec = PetscGetKokkosExecutionSpace();
+  Op                   op;
+  const PetscInt      *iopt = opt ? opt->array : NULL;
+  const PetscInt       M = EQ ? 1 : link->bs / BS, MBS = M * BS;
+  Type                *data = static_cast<Type *>(data_);
+  const Type          *buf  = static_cast<const Type *>(buf_);
+  DeviceExecutionSpace exec = PetscGetKokkosExecutionSpace();
 
   PetscFunctionBegin;
   Kokkos::parallel_for(
@@ -274,11 +273,11 @@ static PetscErrorCode UnpackAndOp(PetscSFLink link, PetscInt count, PetscInt sta
 template <typename Type, class Op, PetscInt BS, PetscInt EQ>
 static PetscErrorCode FetchAndOp(PetscSFLink link, PetscInt count, PetscInt start, PetscSFPackOpt opt, const PetscInt *idx, void *data, void *buf)
 {
-  Op                    op;
-  const PetscInt       *ropt = opt ? opt->array : NULL;
-  const PetscInt        M = EQ ? 1 : link->bs / BS, MBS = M * BS;
-  Type                 *rootdata = static_cast<Type *>(data), *leafbuf = static_cast<Type *>(buf);
-  DeviceExecutionSpace &exec = PetscGetKokkosExecutionSpace();
+  Op                   op;
+  const PetscInt      *ropt = opt ? opt->array : NULL;
+  const PetscInt       M = EQ ? 1 : link->bs / BS, MBS = M * BS;
+  Type                *rootdata = static_cast<Type *>(data), *leafbuf = static_cast<Type *>(buf);
+  DeviceExecutionSpace exec = PetscGetKokkosExecutionSpace();
 
   PetscFunctionBegin;
   Kokkos::parallel_for(
@@ -293,11 +292,11 @@ static PetscErrorCode FetchAndOp(PetscSFLink link, PetscInt count, PetscInt star
 template <typename Type, class Op, PetscInt BS, PetscInt EQ>
 static PetscErrorCode ScatterAndOp(PetscSFLink link, PetscInt count, PetscInt srcStart, PetscSFPackOpt srcOpt, const PetscInt *srcIdx, const void *src_, PetscInt dstStart, PetscSFPackOpt dstOpt, const PetscInt *dstIdx, void *dst_)
 {
-  PetscInt              srcx = 0, srcy = 0, srcX = 0, srcY = 0, dstx = 0, dsty = 0, dstX = 0, dstY = 0;
-  const PetscInt        M = (EQ) ? 1 : link->bs / BS, MBS = M * BS;
-  const Type           *src  = static_cast<const Type *>(src_);
-  Type                 *dst  = static_cast<Type *>(dst_);
-  DeviceExecutionSpace &exec = PetscGetKokkosExecutionSpace();
+  PetscInt             srcx = 0, srcy = 0, srcX = 0, srcY = 0, dstx = 0, dsty = 0, dstX = 0, dstY = 0;
+  const PetscInt       M = (EQ) ? 1 : link->bs / BS, MBS = M * BS;
+  const Type          *src  = static_cast<const Type *>(src_);
+  Type                *dst  = static_cast<Type *>(dst_);
+  DeviceExecutionSpace exec = PetscGetKokkosExecutionSpace();
 
   PetscFunctionBegin;
   /* The 3D shape of source subdomain may be different than that of the destination, which makes it difficult to use CUDA 3D grid and block */
@@ -358,9 +357,9 @@ static PetscErrorCode ScatterAndOp(PetscSFLink link, PetscInt count, PetscInt sr
 template <typename Type, PetscInt BS, PetscInt EQ>
 static PetscErrorCode ScatterAndInsert(PetscSFLink link, PetscInt count, PetscInt srcStart, PetscSFPackOpt srcOpt, const PetscInt *srcIdx, const void *src_, PetscInt dstStart, PetscSFPackOpt dstOpt, const PetscInt *dstIdx, void *dst_)
 {
-  const Type           *src  = static_cast<const Type *>(src_);
-  Type                 *dst  = static_cast<Type *>(dst_);
-  DeviceExecutionSpace &exec = PetscGetKokkosExecutionSpace();
+  const Type          *src  = static_cast<const Type *>(src_);
+  Type                *dst  = static_cast<Type *>(dst_);
+  DeviceExecutionSpace exec = PetscGetKokkosExecutionSpace();
 
   PetscFunctionBegin;
   if (!count) PetscFunctionReturn(PETSC_SUCCESS);
@@ -379,13 +378,13 @@ static PetscErrorCode ScatterAndInsert(PetscSFLink link, PetscInt count, PetscIn
 template <typename Type, class Op, PetscInt BS, PetscInt EQ>
 static PetscErrorCode FetchAndOpLocal(PetscSFLink link, PetscInt count, PetscInt rootstart, PetscSFPackOpt rootopt, const PetscInt *rootidx, void *rootdata_, PetscInt leafstart, PetscSFPackOpt leafopt, const PetscInt *leafidx, const void *leafdata_, void *leafupdate_)
 {
-  Op                    op;
-  const PetscInt        M = (EQ) ? 1 : link->bs / BS, MBS = M * BS;
-  const PetscInt       *ropt     = rootopt ? rootopt->array : NULL;
-  const PetscInt       *lopt     = leafopt ? leafopt->array : NULL;
-  Type                 *rootdata = static_cast<Type *>(rootdata_), *leafupdate = static_cast<Type *>(leafupdate_);
-  const Type           *leafdata = static_cast<const Type *>(leafdata_);
-  DeviceExecutionSpace &exec     = PetscGetKokkosExecutionSpace();
+  Op                   op;
+  const PetscInt       M = (EQ) ? 1 : link->bs / BS, MBS = M * BS;
+  const PetscInt      *ropt     = rootopt ? rootopt->array : NULL;
+  const PetscInt      *lopt     = leafopt ? leafopt->array : NULL;
+  Type                *rootdata = static_cast<Type *>(rootdata_), *leafupdate = static_cast<Type *>(leafupdate_);
+  const Type          *leafdata = static_cast<const Type *>(leafdata_);
+  DeviceExecutionSpace exec     = PetscGetKokkosExecutionSpace();
 
   PetscFunctionBegin;
   Kokkos::parallel_for(
@@ -566,7 +565,8 @@ static PetscErrorCode PetscSFLinkSyncDevice_Kokkos(PetscSFLink PETSC_UNUSED link
 
 static PetscErrorCode PetscSFLinkSyncStream_Kokkos(PetscSFLink PETSC_UNUSED link)
 {
-  DeviceExecutionSpace &exec = PetscGetKokkosExecutionSpace();
+  DeviceExecutionSpace exec = PetscGetKokkosExecutionSpace();
+
   PetscFunctionBegin;
   exec.fence();
   PetscFunctionReturn(PETSC_SUCCESS);
@@ -574,11 +574,12 @@ static PetscErrorCode PetscSFLinkSyncStream_Kokkos(PetscSFLink PETSC_UNUSED link
 
 static PetscErrorCode PetscSFLinkMemcpy_Kokkos(PetscSFLink PETSC_UNUSED link, PetscMemType dstmtype, void *dst, PetscMemType srcmtype, const void *src, size_t n)
 {
-  DeviceExecutionSpace &exec = PetscGetKokkosExecutionSpace();
+  DeviceExecutionSpace exec = PetscGetKokkosExecutionSpace();
 
   PetscFunctionBegin;
   if (!n) PetscFunctionReturn(PETSC_SUCCESS);
-  if (PetscMemTypeHost(dstmtype) && PetscMemTypeHost(srcmtype)) {
+  if (PetscMemTypeHost(dstmtype) && PetscMemTypeHost(srcmtype)) { // H2H
+    PetscCallCXX(exec.fence());                                   // make sure async kernels on src are finished, in case of unified memory as on AMD MI300A.
     PetscCall(PetscMemcpy(dst, src, n));
   } else {
     if (PetscMemTypeDevice(dstmtype) && PetscMemTypeHost(srcmtype)) { // H2D
@@ -592,7 +593,7 @@ static PetscErrorCode PetscSFLinkMemcpy_Kokkos(PetscSFLink PETSC_UNUSED link, Pe
       PetscCallCXX(Kokkos::deep_copy(exec, dbuf, sbuf));
       PetscCallCXX(exec.fence()); // make sure dbuf is ready for use immediately on host
       PetscCall(PetscLogGpuToCpu(n));
-    } else if (PetscMemTypeDevice(dstmtype) && PetscMemTypeDevice(srcmtype)) {
+    } else if (PetscMemTypeDevice(dstmtype) && PetscMemTypeDevice(srcmtype)) { // D2D
       deviceBuffer_t      dbuf(static_cast<char *>(dst), n);
       deviceConstBuffer_t sbuf(static_cast<const char *>(src), n);
       PetscCallCXX(Kokkos::deep_copy(exec, dbuf, sbuf));
@@ -607,7 +608,7 @@ PetscErrorCode PetscSFMalloc_Kokkos(PetscMemType mtype, size_t size, void **ptr)
   if (PetscMemTypeHost(mtype)) PetscCall(PetscMalloc(size, ptr));
   else if (PetscMemTypeDevice(mtype)) {
     if (!PetscKokkosInitialized) PetscCall(PetscKokkosInitializeCheck());
-    PetscCallCXX(*ptr = Kokkos::kokkos_malloc<DeviceMemorySpace>(size));
+    PetscCallCXX(*ptr = Kokkos::kokkos_malloc<DefaultMemorySpace>(size));
   } else SETERRQ(PETSC_COMM_SELF, PETSC_ERR_ARG_WRONG, "Wrong PetscMemType %d", (int)mtype);
   PetscFunctionReturn(PETSC_SUCCESS);
 }
@@ -617,7 +618,7 @@ PetscErrorCode PetscSFFree_Kokkos(PetscMemType mtype, void *ptr)
   PetscFunctionBegin;
   if (PetscMemTypeHost(mtype)) PetscCall(PetscFree(ptr));
   else if (PetscMemTypeDevice(mtype)) {
-    PetscCallCXX(Kokkos::kokkos_free<DeviceMemorySpace>(ptr));
+    PetscCallCXX(Kokkos::kokkos_free<DefaultMemorySpace>(ptr));
   } else SETERRQ(PETSC_COMM_SELF, PETSC_ERR_ARG_WRONG, "Wrong PetscMemType %d", (int)mtype);
   PetscFunctionReturn(PETSC_SUCCESS);
 }
@@ -736,9 +737,9 @@ PetscErrorCode PetscSFLinkSetUp_Kokkos(PetscSF PETSC_UNUSED sf, PetscSFLink link
       PackInit_ComplexType<Kokkos::complex<PetscReal>, 1, 0>(link);
 #endif
   } else {
-    MPI_Aint lb, nbyte;
-    PetscCallMPI(MPI_Type_get_extent(unit, &lb, &nbyte));
-    PetscCheck(lb == 0, PETSC_COMM_SELF, PETSC_ERR_SUP, "Datatype with nonzero lower bound %ld", (long)lb);
+    MPI_Aint nbyte;
+
+    PetscCall(PetscSFGetDatatypeSize_Internal(PETSC_COMM_SELF, unit, &nbyte));
     if (nbyte % sizeof(int)) { /* If the type size is not multiple of int */
 #if !defined(PETSC_HAVE_DEVICE)
       if (nbyte == 4) PackInit_DumbType<char, 4, 1>(link);
@@ -750,7 +751,7 @@ PetscErrorCode PetscSFLinkSetUp_Kokkos(PetscSF PETSC_UNUSED sf, PetscSFLink link
 #endif
         PackInit_DumbType<char, 1, 0>(link);
     } else {
-      nInt = nbyte / sizeof(int);
+      PetscCall(PetscIntCast(nbyte / sizeof(int), &nInt));
 #if !defined(PETSC_HAVE_DEVICE)
       if (nInt == 8) PackInit_DumbType<int, 8, 1>(link);
       else if (nInt % 8 == 0) PackInit_DumbType<int, 8, 0>(link);
